/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"


// current match cycle
char *ac_match[MAX_LIST_ITEMS];
int ac_nummatch = 0;
int ac_current = 0;

itemlist_t ac_maps = {NULL, 0, NULL};
itemlist_t ac_texmode = {NULL, 0, NULL};
itemlist_t ac_give = {NULL, 0, NULL};
itemlist_t ac_sky = {NULL, 0, NULL};
itemlist_t ac_cfg = {NULL, 0, NULL};
itemlist_t ac_load = {NULL, 0, NULL};


// setup all of our autocompletes so that we don't need a separate scanning function for each
typedef struct autocompnames_s
{
	char *name;
	itemlist_t *ac;
} autocompnames_t;

autocompnames_t autocompnames[] = {
	{"map ", &ac_maps},
	{"gl_texturemode ", &ac_texmode},
	{"give ", &ac_give},
	{"sky ", &ac_sky},
	{"loadsky ", &ac_sky},
	{"exec ", &ac_cfg},
	{"load ", &ac_load},
	{"save ", &ac_load},
	{NULL, NULL}
};


void Cmd_ClearAutoComplete (itemlist_t *ac)
{
	// using a Win32 heap here instead of malloc and free because we can throw everything away with a simple HeapDestroy
	if (ac->heap)
	{
		HeapDestroy (ac->heap);
		ac->heap = NULL;
	}

	memset (ac, 0, sizeof (itemlist_t));
	ac->numitems = 0;
}


void Cmd_CopyStringToAutoComplete (itemlist_t *ac, char *name)
{
	if (ac->numitems < MAX_LIST_ITEMS)
	{
		if (!ac->heap)
		{
			// using a Win32 heap here instead of malloc and free because we can throw everything away with a simple HeapDestroy
			ac->heap = HeapCreate (0, 0, 0);
		}

		if (!ac->items)
		{
			// alloc an initial heap of the max number of items we are going to support
			ac->items = (char **) HeapAlloc (ac->heap, HEAP_ZERO_MEMORY, MAX_LIST_ITEMS * sizeof (char *));
		}

		// don't add items more than once
		for (int i = 0; i < ac->numitems; i++)
			if (!strcmp (name, ac->items[i]))
				return;

		ac->items[ac->numitems] = (char *) HeapAlloc (ac->heap, HEAP_ZERO_MEMORY, strlen (name) + 1);
		strcpy (ac->items[ac->numitems], name);
		ac->numitems++;
	}
}


void Cmd_CopyFileToAutoComplete (itemlist_t *ac, char *name, qboolean removeext)
{
	char *shortname = (char *) Scratch_Alloc ();

	// copy it over initially in case there is no path
	strcpy (shortname, name);

	// remove the path
	for (int i = strlen (shortname); i > 0; i--)
	{
		if (shortname[i] == '/' || shortname[i] == '\\')
		{
			// copy it over
			shortname = &shortname[i + 1];
			break;
		}
	}

	// optionally remove the extension
	if (removeext)
	{
		for (int i = strlen (shortname); i > 0; i--)
		{
			if (shortname[i] == '.')
			{
				shortname[i] = 0;
				break;
			}
		}
	}

	// and do the copy
	Cmd_CopyStringToAutoComplete (ac, shortname);
}


void Cmd_AutoCompleteCallback (char *name)
{
	char *ext = NULL;

	for (int i = strlen (name); i > 0; i--)
	{
		if (name[i] == '.')
		{
			ext = &name[i];
			break;
		}
	}

	if (!ext) return;

	// add all the file types that we wish to use
	if (!stricmp (ext, ".bsp"))
	{
		Cmd_CopyFileToAutoComplete (&ac_maps, name, true);
		return;
	}

	if (!stricmp (ext, ".dm2"))
	{
		Cmd_CopyFileToAutoComplete (&ac_maps, name, false);
		return;
	}

	if (!stricmp (ext, ".cin"))
	{
		Cmd_CopyFileToAutoComplete (&ac_maps, name, false);
		return;
	}

	if (strstr (name, "env/") && (!stricmp (ext, ".tga") || !stricmp (ext, ".pcx")))
	{
		// copy it off separately because we must remove the suffix (which will also remove the ext)
		char skyname[256];

		strcpy (skyname, name);

		for (int i = strlen (skyname); i > 3; i--)
		{
			if (skyname[i] == '.')
			{
				// remove the suffix
				skyname[i - 2] = 0;
				break;
			}
		}

		// no need to remove the ext because it's already been removed
		Cmd_CopyFileToAutoComplete (&ac_sky, skyname, false);
		return;
	}

	if (!stricmp (ext, ".cfg"))
	{
		// exec needs the extension
		Cmd_CopyFileToAutoComplete (&ac_cfg, name, false);
		return;
	}
}


int Cmd_AutoCompleteCompare (const void *a, const void *b)
{
	return strcmp (*(const char **) a, *(const char **) b);
}


void Cmd_FinalizeAutoComplete (itemlist_t *ac)
{
	if (ac->numitems > 1)
		qsort (ac->items, ac->numitems, sizeof (char *), Cmd_AutoCompleteCompare);
	else if (!ac->numitems)
	{
		Cmd_ClearAutoComplete (ac);
		return;
	}

	// now allocate the items for real and copy them over; allocating one extra for NULL terminating the list in case anything needs it
	// doing it this way, we can pass the char** of an autocomp list to the menus as a spincontrol
	char **copied = (char **) HeapAlloc (ac->heap, HEAP_ZERO_MEMORY, (ac->numitems + 1) * sizeof (char *));

	memcpy (copied, ac->items, ac->numitems * sizeof (char *));
	HeapFree (ac->heap, 0, ac->items);
	ac->items = copied;

	// compact the heap to defrag address space
	HeapCompact (ac->heap, 0);
}


void Cmd_SetupAutoCompleteNames (itemlist_t *ac, char *names[])
{
	Cmd_ClearAutoComplete (ac);

	for (int i = 0;; i++)
	{
		if (!names[i]) break;
		Cmd_CopyStringToAutoComplete (ac, names[i]);
	}

	Cmd_FinalizeAutoComplete (ac);
}


void Cmd_SetupSaveGamesAutoComplete (void)
{
	// clear down autocomplete lists
	Cmd_ClearAutoComplete (&ac_load);

	// enumerate saves and add to the list
	FS_EnumSaves (&ac_load);

	// finalize the lists
	Cmd_FinalizeAutoComplete (&ac_load);
}


void Cmd_InitAutoComplete (void)
{
	static char *texmode_names[] = {
		"GL_NEAREST",
		"GL_LINEAR",
		"GL_NEAREST_MIPMAP_NEAREST",
		"GL_LINEAR_MIPMAP_NEAREST",
		"GL_NEAREST_MIPMAP_LINEAR",
		"GL_LINEAR_MIPMAP_LINEAR",
		NULL
	};

	static char *give_names[] = {
		"adrenaline",
		"airstrike marker",
		"all",
		"ammo pack",
		"ammo",
		"ancient head",
		"armor shard",
		"armor",
		"bandolier",
		"bfg10k",
		"blue key",
		"bullets",
		"cells",
		"chaingun",
		"combat armor",
		"commander's head",
		"data cd",
		"data spinner",
		"environment suit",
		"grenade launcher",
		"grenades",
		"health",
		"hyperblaster",
		"invulnerability",
		"jacket armor",
		"machinegun",
		"power cube",
		"power screen",
		"power shield",
		"pyramid key",
		"quad damage",
		"railgun",
		"rebreather",
		"red key",
		"rocket launcher",
		"rockets",
		"security pass",
		"shells",
		"shotgun",
		"silencer",
		"slugs",
		"super shotgun",
		"weapons",
		NULL
	};

	// clear down autocomplete lists
	Cmd_ClearAutoComplete (&ac_maps);
	Cmd_ClearAutoComplete (&ac_sky);
	Cmd_ClearAutoComplete (&ac_cfg);

	// enumerate filesystem content to add to the lists
	FS_EnumContent (Cmd_AutoCompleteCallback);

	// finalize the lists
	Cmd_FinalizeAutoComplete (&ac_maps);
	Cmd_FinalizeAutoComplete (&ac_sky);
	Cmd_FinalizeAutoComplete (&ac_cfg);

	// set up the other autocomplete lists
	Cmd_SetupAutoCompleteNames (&ac_give, give_names);
	Cmd_SetupAutoCompleteNames (&ac_texmode, texmode_names);

	// set up the save games list which is done outside of the other FS content because it needs to enumerate folders, not files
	// (also because it needs to be re-done any time a game is saved which may change the list)
	Cmd_SetupSaveGamesAutoComplete ();
}


void Cmd_EndAutoComplete (void)
{
	ac_current = 0;
	ac_nummatch = 0;
}


void Cmd_BeginAutoComplete (itemlist_t *ac, char *comp)
{
	// eand any current autocomplete cycle
	Cmd_EndAutoComplete ();

	// store out all the matches when we begin an autocomplete cycle
	for (int i = 0; i < ac->numitems; i++)
	{
		if (strnicmp (ac->items[i], comp, strlen (comp))) continue;

		ac_match[ac_nummatch] = ac->items[i];
		ac_nummatch++;
	}
}


char *Cmd_RunAutoComplete (char *comp)
{
	for (int i = 0;; i++)
	{
		// stop at an empty slot
		if (!autocompnames[i].name) break;
		if (!autocompnames[i].ac) break;

		int len = strlen (autocompnames[i].name);

		if (!strnicmp (comp, autocompnames[i].name, len))
		{
			static char found[256];

			if (!ac_nummatch) Cmd_BeginAutoComplete (autocompnames[i].ac, &comp[len]);
			if (!ac_nummatch) break;

			// copy off the current match
			sprintf (found, "%s%s", autocompnames[i].name, ac_match[ac_current]);

			// go to the next
			ac_current = (ac_current + 1) % ac_nummatch;

			// and return what we got
			return found;
		}
	}

	// not found at all
	Cmd_EndAutoComplete ();
	return NULL;
}


