/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"

#define _TINYDIR_MALLOC Z_Alloc
#define _TINYDIR_FREE Z_Free

#include "tinydir.h"

void Cmd_InitAutoComplete (void);

// to do - memory mapped files


// if a packfile directory differs from this, it is assumed to be hacked
#define	PAK0_CHECKSUM	0x40e614e0		// Full version
//#define	PAK0_CHECKSUM	0xb2c6d7ea		// demo
//#define	PAK0_CHECKSUM	0x78e135c		// OEM

/*
=============================================================================

QUAKE FILESYSTEM

=============================================================================
*/

// in memory
typedef struct packfile_s
{
	char	name[MAX_QPATH];
	int		filepos, filelen;
} packfile_t;

typedef struct pack_s
{
	char	filename[MAX_OSPATH];
	int		numfiles;
	packfile_t	*files;
} pack_t;

char	fs_gamedir[MAX_OSPATH];
cvar_t	*fs_gamedirvar;

typedef struct searchpath_s
{
	char	filename[MAX_OSPATH];
	pack_t	*pack;		// only one of filename / pack will be used
	struct searchpath_s *next;
} searchpath_t;

searchpath_t	*fs_searchpaths;
searchpath_t	*fs_base_searchpaths;	// without gamedirs


/*
All of Quake's data access is through a hierchal file system, but the contents of the file system can be transparently merged from several sources.

The "base directory" is the path to the directory holding the quake.exe and all game directories.  The sys_* files pass this to host_init in
quakeparms_t->basedir.  This can be overridden with the "-basedir" command line parm to allow code debugging in a different directory.  The base
directory is only used during filesystem initialization.

The "game directory" is the first tree on the search path and directory that all generated files (savegames, screenshots, demos, config files)
will be saved to.  This can be overridden with the "-game" command line parameter.  The game directory can never be changed while quake is
executing.  This is a precacution against having a malicious server instruct clients to write files over areas they shouldn't.
*/


/*
================
FS_FileLength
================
*/
int FS_FileLength (qfile_t *f)
{
	if (f->handle == INVALID_HANDLE_VALUE)
		return -1;
	else return GetFileSize (f->handle, NULL);
}


/*
============
FS_CreatePath

Creates any directories needed to store the given filename
============
*/
void FS_CreatePath (char *path)
{
	for (char *ofs = path + 1; *ofs; ofs++)
	{
		if (*ofs == '/')
		{
			// create the directory
			*ofs = 0;
			Sys_Mkdir (path);
			*ofs = '/';
		}
	}
}


/*
==============
FS_FCloseFile

For some reason, other dll's can't just call fclose()
on files returned by FS_FOpenFile...

(presumably because of memory allocation boundary rules)
safely closes the file and NULLs the handle so that any future attempts to access it are errors
==============
*/
void FS_FCloseFile (qfile_t *f)
{
	if (f->handle != INVALID_HANDLE_VALUE)
	{
		CloseHandle (f->handle);
		memset (f, 0, sizeof (qfile_t));
		f->handle = INVALID_HANDLE_VALUE;
	}
}


// RAFAEL
int	Developer_searchpath (int who)
{
	for (searchpath_t *search = fs_searchpaths; search; search = search->next)
	{
		if (strstr (search->filename, "xatrix")) return 1;
		if (strstr (search->filename, "rogue")) return 2;
	}

	return (0);
}


packfile_t *FS_FindFileInPAK (packfile_t *files, int numfiles, char *filename)
{
	// because q2 pack files are so much bigger and have so many more files we binary search them
	// instead of doing a linear search; we could have used bsearch but we'd have to construct a
	// packfile_t to use as a key - this way we can just use a file name instead
	assert (files);
	assert (filename);
	assert (numfiles >= 0);

	// to do - we could potentially short-circuit some of this by defining a range to search in based on the first char of the filename
	int imin = 0;
	int imax = numfiles - 1;

	for (;;)
	{
		if (imax < imin) break;

		// some pack files have mixed-case file names so do a lower-case comparison
		int imid = (imax + imin) >> 1;
		int comp = _stricmp (files[imid].name, filename);

		if (comp < 0)
			imin = imid + 1;
		else if (comp > 0)
			imax = imid - 1;
		else return &files[imid];
	}

	// not found
	return NULL;
}


// does a direct fopen on a file in the filesystem for when we want to bypass the search paths (e.g for a save game)
qboolean FS_FOpenRead (char *filename, qfile_t *f)
{
	f->filelen = -1;

	// ensure that the qfile_t was properly initialized
	assert (f->handle == INVALID_HANDLE_VALUE);

	if ((f->handle = CreateFile (filename, FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL)) != INVALID_HANDLE_VALUE)
	{
		f->filelen = FS_FileLength (f);
		return true;
	}
	else return false;
}


qboolean FS_FOpenWrite (char *filename, qfile_t *f)
{
	f->filelen = -1;

	// ensure that the qfile_t was properly initialized
	assert (f->handle == INVALID_HANDLE_VALUE);

	if ((f->handle = CreateFile (filename, FILE_WRITE_DATA, 0, NULL, CREATE_ALWAYS, 0, NULL)) != INVALID_HANDLE_VALUE)
		return true;
	else return false;
}


qboolean FS_FOpenAppend (char *filename, qfile_t *f)
{
	f->filelen = -1;

	// ensure that the qfile_t was properly initialized
	assert (f->handle == INVALID_HANDLE_VALUE);

	if ((f->handle = CreateFile (filename, FILE_APPEND_DATA, 0, NULL, OPEN_ALWAYS, 0, NULL)) != INVALID_HANDLE_VALUE)
	{
		f->filelen = FS_FileLength (f);
		return true;
	}
	else return false;
}


/*
===========
FS_FOpenFile

Finds the file in the search path.
returns filesize and an open FILE *
Used for streaming data out of either a pak file or
a seperate file.
===========
*/
int file_from_pak = 0;

int FS_FOpenFile (char *filename, qfile_t *file)
{
	searchpath_t	*search;
	char			netpath[MAX_OSPATH];
	pack_t			*pak;
	int				i;

	// to do: memory-map the files and leave PAK files open so that we can always access them quickly
	// and don't need to be constantly copying data from one buffer to another
	file_from_pak = 0;

	// ensure that the qfile_t was properly initialized
	assert (file->handle == INVALID_HANDLE_VALUE);

	// clear the file
	memset (file, 0, sizeof (qfile_t));
	file->handle = INVALID_HANDLE_VALUE;

	// search through the path, one element at a time
	for (search = fs_searchpaths; search; search = search->next)
	{
		// is the element a pak file?
		if (search->pack)
		{
			packfile_t *pf = NULL;

			// look through all the pak file elements
			pak = search->pack;

			if ((pf = FS_FindFileInPAK (pak->files, pak->numfiles, filename)) != NULL)
			{
				// found it!
				file_from_pak = 1;
				Com_DPrintf ("PackFile: %s : %s\n", pak->filename, filename);

				// open a new file on the pakfile
				if (!FS_FOpenRead (pak->filename, file))
					Com_Error (ERR_FATAL, "Couldn't reopen %s", pak->filename);

				file->filelen = pf->filelen;

				// seek to the start of the file
				FS_Seek (file, pf->filepos, FILE_BEGIN);

				return pf->filelen;
			}
		}
		else
		{
			// check a file in the directory tree
			Com_sprintf (netpath, sizeof (netpath), "%s/%s", search->filename, filename);

			if (!FS_FOpenRead (netpath, file))
				continue;

			Com_DPrintf ("FindFile: %s\n", netpath);
			file->filelen = FS_FileLength (file);
			return file->filelen;
		}
	}

	Com_DPrintf ("FindFile: can't find %s\n", filename);

	// clear the file
	memset (file, 0, sizeof (qfile_t));
	file->handle = INVALID_HANDLE_VALUE;
	return -1;
}



/*
=================
FS_ReadFile

Properly handles partial reads
=================
*/
int FS_Read (void *buffer, int len, qfile_t *f)
{
	DWORD readlen = 0;

	// because there is no progress bar thing we just do a straight-up full read instead of a chunked read
	// fread returns 1 here and so should we (to maintain consistency)
	if (f->handle == INVALID_HANDLE_VALUE)
		return 0;
	else if (ReadFile (f->handle, buffer, len, &readlen, NULL))
	{
		if (readlen != len)
			return 0;
		else return 1;
	}
	else return 0;
}

int FS_Seek (qfile_t *f, int offset, DWORD method)
{
	if (f->handle == INVALID_HANDLE_VALUE)
		return 0;
	else return SetFilePointer (f->handle, offset, NULL, method);
}


int FS_Write (void *buffer, int len, qfile_t *f)
{
	DWORD writelen = 0;

	// fwrite returns 1 here and so should we (to maintain consistency)
	if (f->handle == INVALID_HANDLE_VALUE)
		return 0;
	else if (WriteFile (f->handle, buffer, len, &writelen, NULL))
	{
		if (writelen != len)
			return 0;
		else return 1;
	}
	else return 0;
}

qboolean FS_FileIsOpen (qfile_t *f)
{
	// !!!!! be very careful of where you call this from ensuring that the qfile_t is properly initialized first !!!!!
	if (f->handle == INVALID_HANDLE_VALUE)
		return false;
	else return true;
}


qboolean FS_FileExists (char *name)
{
	DECLARE_FILE (f);

	if (FS_FOpenRead (name, &f))
	{
		FS_FCloseFile (&f);
		return true;
	}
	else return false;
}


/*
================
FS_CopyFile
================
*/
void FS_CopyFile (char *src, char *dst)
{
	CopyFile (src, dst, FALSE);
}


/*
============
FS_LoadFile

Filename are reletive to the quake search path
a null buffer will just return the file length without loading
============
*/
int FS_LoadFile (char *path, void **buffer)
{
	DECLARE_FILE (f);
	byte	*buf;
	int		len;

	buf = NULL;	// quiet compiler warning

	// look for it in the filesystem or pack files
	len = FS_FOpenFile (path, &f);

	if (f.handle == INVALID_HANDLE_VALUE)
	{
		if (buffer)
			*buffer = NULL;

		return -1;
	}

	if (!buffer)
	{
		FS_FCloseFile (&f);
		return len;
	}

	buf = Z_Alloc (len);
	*buffer = buf;

	FS_Read (buf, len, &f);
	FS_FCloseFile (&f);

	return len;
}


/*
=============
FS_FreeFile
=============
*/
void FS_FreeFile (void *buffer)
{
	Z_Free (buffer);
}

int FS_PackFileCompFunc (packfile_t *p1, packfile_t *p2)
{
	// some pack files have mixed-case file names so do a lower-case comparison
	return _stricmp (p1->name, p2->name);
}


/*
=================
FS_LoadPackFile

Takes an explicit (not game tree related) path to a pak file.

Loads the header and directory, adding the files at the beginning
of the list so they override previous pack files.
=================
*/
pack_t *FS_LoadPackFile (char *packfile)
{
	dpackheader_t	header;
	int				i;
	packfile_t		*newfiles;
	int				numpackfiles;
	pack_t			*pack;
	DECLARE_FILE	(packhandle);
	dpackfile_t		*info = (dpackfile_t *) Scratch_Alloc ();
	unsigned		checksum;

	if (!FS_FOpenRead (packfile, &packhandle))
		return NULL;

	FS_Read (&header, sizeof (header), &packhandle);

	if (LittleLong (header.ident) != IDPAKHEADER)
		Com_Error (ERR_FATAL, "%s is not a packfile", packfile);

	header.dirofs = LittleLong (header.dirofs);
	header.dirlen = LittleLong (header.dirlen);

	numpackfiles = header.dirlen / sizeof (dpackfile_t);

	if (numpackfiles > MAX_FILES_IN_PACK)
		Com_Error (ERR_FATAL, "%s has %i files", packfile, numpackfiles);

	newfiles = Z_Alloc (numpackfiles * sizeof (packfile_t));

	FS_Seek (&packhandle, header.dirofs, FILE_BEGIN);
	FS_Read (info, header.dirlen, &packhandle);
	FS_FCloseFile (&packhandle);

	// crc the directory to check for modifications
	checksum = Com_BlockChecksum ((void *) info, header.dirlen);

	// parse the directory
	for (i = 0; i < numpackfiles; i++)
	{
		strcpy (newfiles[i].name, info[i].name);
		newfiles[i].filepos = LittleLong (info[i].filepos);
		newfiles[i].filelen = LittleLong (info[i].filelen);
	}

	pack = Z_Alloc (sizeof (pack_t));
	strcpy (pack->filename, packfile);
	pack->numfiles = numpackfiles;
	pack->files = newfiles;

	// sort the pack files by name so that we can perform a binary search on them
	qsort (pack->files, pack->numfiles, sizeof (packfile_t), (sortfunc_t) FS_PackFileCompFunc);

	Com_Printf ("Added packfile %s (%i files)\n", packfile, numpackfiles);
	return pack;
}


/*
================
FS_AddGameDirectory

Sets fs_gamedir, adds the directory to the head of the path,
then loads and adds pak1.pak pak2.pak ...
================
*/
void FS_AddGameDirectory (char *dir)
{
	int				i;
	searchpath_t	*search;
	pack_t			*pak;
	char			pakfile[MAX_OSPATH];

	strcpy (fs_gamedir, dir);

	// add the directory to the search path
	search = Z_Alloc (sizeof (searchpath_t));
	strcpy (search->filename, dir);
	search->next = fs_searchpaths;
	fs_searchpaths = search;

	// add any pak files in the format pak0.pak pak1.pak, ...
	for (i = 0; i < 10; i++)
	{
		Com_sprintf (pakfile, sizeof (pakfile), "%s/pak%i.pak", dir, i);
		pak = FS_LoadPackFile (pakfile);

		if (!pak)
			continue;

		search = Z_Alloc (sizeof (searchpath_t));
		search->pack = pak;
		search->next = fs_searchpaths;
		fs_searchpaths = search;
	}
}


/*
============
FS_Gamedir

Called to find where to write a file (demos, savegames, etc)
============
*/
char *FS_Gamedir (void)
{
	return fs_gamedir;
}

/*
=============
FS_ExecAutoexec
=============
*/
void FS_ExecAutoexec (void)
{
	char *dir = Cvar_VariableString ("gamedir");
	char name[MAX_QPATH];

	if (*dir)
		Com_sprintf (name, sizeof (name), "./%s/autoexec.cfg", dir);
	else Com_sprintf (name, sizeof (name), "./%s/autoexec.cfg", BASEDIRNAME);

	if (Sys_FindFirst (name, 0, SFF_SUBDIR | SFF_HIDDEN | SFF_SYSTEM))
		Cbuf_AddText ("exec autoexec.cfg\n");

	Sys_FindClose ();
}


/*
================
FS_SetGamedir

Sets the gamedir and path to a different directory.
================
*/
void FS_SetGamedir (char *dir)
{
	searchpath_t	*next;

	if (strstr (dir, "..") || strstr (dir, "/") || strstr (dir, "\\") || strstr (dir, ":"))
	{
		Com_Printf ("Gamedir should be a single filename, not a path\n");
		return;
	}

	// free up any current game dir info
	while (fs_searchpaths != fs_base_searchpaths)
	{
		if (fs_searchpaths->pack)
		{
			Z_Free (fs_searchpaths->pack->files);
			Z_Free (fs_searchpaths->pack);
		}

		next = fs_searchpaths->next;
		Z_Free (fs_searchpaths);
		fs_searchpaths = next;
	}

	// flush all data, so it will be forced to reload
	if (dedicated && !dedicated->value)
	{
		Cbuf_AddText ("snd_restart\n");
		re->UnloadGfx ();
	}

	Com_sprintf (fs_gamedir, sizeof (fs_gamedir), "./%s", dir);

	if (!strcmp (dir, BASEDIRNAME) || (*dir == 0))
	{
		Cvar_FullSet ("gamedir", "", CVAR_SERVERINFO | CVAR_NOSET);
		Cvar_FullSet ("game", "", CVAR_LATCH | CVAR_SERVERINFO);
	}
	else
	{
		Cvar_FullSet ("gamedir", dir, CVAR_SERVERINFO | CVAR_NOSET);
		FS_AddGameDirectory (va ("./%s", dir));
	}

	// reload anything that needs to be reloaded by hand
	if (dedicated && !dedicated->value)
	{
		re->LoadGfx ();
	}

	// init map autocompletion by reading the maps.lst file into the autocompletion list
	Cmd_InitAutoComplete ();
}


/*
================
FS_ListFiles
================
*/
char **FS_ListFiles (char *findname, int *numfiles, unsigned musthave, unsigned canthave)
{
	char *s;
	int nfiles = 0;
	char **list = 0;

	s = Sys_FindFirst (findname, musthave, canthave);

	while (s)
	{
		if (s[strlen (s) - 1] != '.')
			nfiles++;

		s = Sys_FindNext (musthave, canthave);
	}

	Sys_FindClose ();

	if (!nfiles)
		return NULL;

	nfiles++; // add space for a guard
	*numfiles = nfiles;

	list = (char **) HeapAlloc (GetProcessHeap (), HEAP_ZERO_MEMORY, sizeof (char *) * nfiles);

	s = Sys_FindFirst (findname, musthave, canthave);
	nfiles = 0;

	while (s)
	{
		if (s[strlen (s) - 1] != '.')
		{
			list[nfiles] = (char *) HeapAlloc (GetProcessHeap (), HEAP_ZERO_MEMORY, strlen (s) + 1);
			strcpy (list[nfiles], s);
			strlwr (list[nfiles]);
			nfiles++;
		}

		s = Sys_FindNext (musthave, canthave);
	}

	Sys_FindClose ();

	return list;
}


/*
===============
FS_Dir_f
===============
*/
void FS_Dir_f (void)
{
	char	findname[1024];
	char	wildcard[1024] = "*.*";
	char	**dirnames;
	int		ndirs;

	if (Cmd_Argc () != 1)
	{
		strcpy (wildcard, Cmd_Argv (1));
	}

	// FS_NextPath will return fs_gamedir twice; once for the NULL starting path, and once for when it actually hits the gamedir
	for (searchpath_t *s = fs_searchpaths; s; s = s->next)
	{
		if (s->pack) continue;

		char *path = s->filename;
		char *tmp = findname;

		Com_sprintf (findname, sizeof (findname), "%s/%s", path, wildcard);

		while (*tmp != 0)
		{
			if (*tmp == '\\')
				*tmp = '/';

			tmp++;
		}

		Com_Printf ("Directory of %s\n", findname);
		Com_Printf ("----\n");

		if ((dirnames = FS_ListFiles (findname, &ndirs, 0, 0)) != 0)
		{
			int i;

			for (i = 0; i < ndirs - 1; i++)
			{
				if (strrchr (dirnames[i], '/'))
					Com_Printf ("%s\n", strrchr (dirnames[i], '/') + 1);
				else Com_Printf ("%s\n", dirnames[i]);

				HeapFree (GetProcessHeap (), 0, dirnames[i]);
			}

			HeapFree (GetProcessHeap (), 0, dirnames);
		}

		Com_Printf ("\n");
	}
}


/*
============
FS_Path_f

============
*/
void FS_Path_f (void)
{
	searchpath_t	*s;

	Com_Printf ("Current search path:\n");

	for (s = fs_searchpaths; s; s = s->next)
	{
		if (s == fs_base_searchpaths)
			Com_Printf ("----------\n");

		if (s->pack)
			Com_Printf ("%s (%i files)\n", s->pack->filename, s->pack->numfiles);
		else Com_Printf ("%s\n", s->filename);
	}
}

/*
================
FS_NextPath

Allows enumerating all of the directories in the search path
================
*/
char *FS_NextPath (char *prevpath)
{
	if (!prevpath)
		return fs_gamedir;

	char *prev = fs_gamedir;

	for (searchpath_t *s = fs_searchpaths; s; s = s->next)
	{
		if (s->pack)
			continue;

		if (prevpath == prev)
			return s->filename;

		prev = s->filename;
	}

	return NULL;
}


/*
================
FS_InitFilesystem
================
*/
void FS_InitFilesystem (void)
{
	Cmd_AddCommand ("path", FS_Path_f);
	Cmd_AddCommand ("dir", FS_Dir_f);

	// start up with baseq2 by default
	FS_AddGameDirectory ("./"BASEDIRNAME);

	// any set gamedirs will be freed up to here
	fs_base_searchpaths = fs_searchpaths;

	// check for game override
	fs_gamedirvar = Cvar_Get ("game", "", CVAR_LATCH | CVAR_SERVERINFO);

	if (fs_gamedirvar->string[0])
		FS_SetGamedir (fs_gamedirvar->string);

	// init map autocompletion by reading the maps.lst file into the autocompletion list
	// (this will be done twice if started with +set but that's OK
	Cmd_InitAutoComplete ();
}


void FS_RecursiveDirectoryTraverse (char *dirname, void (*Callback) (char *))
{
	tinydir_dir dir;

	tinydir_open (&dir, dirname);

	while (dir.has_next)
	{
		tinydir_file file;

		tinydir_readfile (&dir, &file);

		// handle the file appropriately
		if (!strcmp (file.name, "."))
			; // skip this dir
		else if (!strcmp (file.name, ".."))
			; // skip this dir
		else if (file.is_dir)
			FS_RecursiveDirectoryTraverse (file.path, Callback);
		else Callback (file.path);

		tinydir_next (&dir);
	}

	tinydir_close (&dir);
}


void FS_EnumContent (void (*Callback) (char *))
{
	for (searchpath_t *s = fs_searchpaths; s; s = s->next)
	{
		if (s->pack)
		{
			pack_t *pak = s->pack;

			for (int i = 0; i < pak->numfiles; i++)
			{
				packfile_t *file = &pak->files[i];
				Callback (file->name);
			}
		}
		else FS_RecursiveDirectoryTraverse (s->filename, Callback);
	}
}


void FS_EnumSaves (itemlist_t *ac)
{
	for (searchpath_t *s = fs_searchpaths; s; s = s->next)
	{
		// no saves in PAK files
		if (s->pack) continue;

		tinydir_dir dir;

		tinydir_open (&dir, va ("%s/save", s->filename));

		while (dir.has_next)
		{
			tinydir_file file;

			// if no saves dir has been created yet, bail out!
			if (tinydir_readfile (&dir, &file) == -1) break;

			// handle the file appropriately
			if (!strcmp (file.name, "."))
				; // skip this dir
			else if (!strcmp (file.name, ".."))
				; // skip this dir
			else if (!file.is_dir)
				; // skip this file
			else Cmd_CopyStringToAutoComplete (ac, file.name);

			tinydir_next (&dir);
		}

		tinydir_close (&dir);

		// the first non-PAK location is where the current gamedir saves are going to be
		break;
	}
}

