/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// models.c -- model loading and caching

#include "gl_local.h"

extern model_t	*loadmodel;
extern int		modfilelen;

extern model_t	mod_known[];
extern model_t	mod_inline[];

/*
===============================================================================

BRUSHMODEL LOADING

===============================================================================
*/


/*
=================
Mod_LoadLighting
=================
*/
void Mod_LoadLighting (lump_t *l, byte *base)
{
	if (!l->filelen)
	{
		loadmodel->lightdata = NULL;
		return;
	}

	loadmodel->lightdata = ri->Hunk_Alloc (l->filelen);
	memcpy (loadmodel->lightdata, base, l->filelen);
}


/*
=================
Mod_LoadVisibility
=================
*/
void Mod_LoadVisibility (lump_t *l, byte *base)
{
	int		i;

	if (!l->filelen)
	{
		loadmodel->vis = NULL;
		return;
	}

	loadmodel->vis = ri->Hunk_Alloc (l->filelen);
	memcpy (loadmodel->vis, base, l->filelen);

	loadmodel->vis->numclusters = LittleLong (loadmodel->vis->numclusters);

	for (i = 0; i < loadmodel->vis->numclusters; i++)
	{
		loadmodel->vis->bitofs[i][0] = LittleLong (loadmodel->vis->bitofs[i][0]);
		loadmodel->vis->bitofs[i][1] = LittleLong (loadmodel->vis->bitofs[i][1]);
	}
}


/*
=================
RadiusFromBounds
=================
*/
float RadiusFromBounds (vec3_t mins, vec3_t maxs)
{
	int		i;
	vec3_t	corner;

	for (i = 0; i < 3; i++)
	{
		corner[i] = fabs (mins[i]) > fabs (maxs[i]) ? fabs (mins[i]) : fabs (maxs[i]);
	}

	return VectorLength (corner);
}


/*
=================
Mod_LoadSubmodels
=================
*/
void Mod_LoadSubmodels (lump_t *l, byte *base)
{
	dmodel_t	*in;
	mmodel_t	*out;
	int			i, j, count;

	in = (void *) base;

	if (l->filelen % sizeof (*in))
		ri->Sys_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*in);
	out = ri->Hunk_Alloc (count * sizeof (*out));

	loadmodel->submodels = out;
	loadmodel->numsubmodels = count;

	for (i = 0; i < count; i++, in++, out++)
	{
		for (j = 0; j < 3; j++)
		{
			// spread the mins / maxs by a pixel
			out->mins[j] = LittleFloat (in->mins[j]) - 1;
			out->maxs[j] = LittleFloat (in->maxs[j]) + 1;
			out->origin[j] = LittleFloat (in->origin[j]);
		}

		out->radius = RadiusFromBounds (out->mins, out->maxs);
		out->headnode = LittleLong (in->headnode);
		out->firstface = LittleLong (in->firstface);
		out->numfaces = LittleLong (in->numfaces);
	}
}


/*
=================
Mod_LoadTexinfo
=================
*/
void Mod_LoadTexinfo (lump_t *l, byte *base)
{
	texinfo_t *in;
	mtexinfo_t *out, *step;
	int 	i, j, count;
	char	name[MAX_QPATH];
	int		next;

	in = (void *) base;

	if (l->filelen % sizeof (*in))
		ri->Sys_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*in);
	out = ri->Hunk_Alloc (count * sizeof (*out));

	loadmodel->texinfo = out;
	loadmodel->numtexinfo = count;

	for (i = 0; i < count; i++, in++, out++)
	{
		for (j = 0; j < 8; j++)
			out->vecs[0][j] = LittleFloat (in->vecs[0][j]);

		out->flags = LittleLong (in->flags);
		next = LittleLong (in->nexttexinfo);

		if (next > 0)
			out->next = loadmodel->texinfo + next;
		else out->next = NULL;

		Com_sprintf (name, sizeof (name), "textures/%s.wal", in->texture);

		out->image = R_FindImage (name, it_wall);

		if (!out->image)
		{
			ri->Con_Printf (PRINT_ALL, "Couldn't load %s\n", name);
			out->image = r_notexture;
		}
	}

	// count animation frames
	for (i = 0; i < count; i++)
	{
		out = &loadmodel->texinfo[i];
		out->numframes = 1;

		for (step = out->next; step && step != out; step = step->next)
			out->numframes++;
	}
}

/*
================
Mod_CalcSurfaceExtents

Fills in s->texturemins[] and s->extents[]
================
*/
void Mod_CalcSurfaceBounds (msurface_t *s, dbsp_t *bsp)
{
	float	*v;

	// clear the bounding box
	Vector3Set (s->mins, 999999, 999999, 999999);
	Vector3Set (s->maxs, -999999, -999999, -999999);

	// get a bounding box for the surf from all of it's verts
	for (int i = 0; i < s->numedges; i++)
	{
		int e = bsp->surfedges[s->firstedge + i];
		float *v;

		if (e >= 0)
			v = bsp->vertexes[bsp->edges[e].v[0]].point;
		else v = bsp->vertexes[bsp->edges[-e].v[1]].point;

		for (int j = 0; j < 3; j++)
		{
			if (v[j] < s->mins[j]) s->mins[j] = v[j];
			if (v[j] > s->maxs[j]) s->maxs[j] = v[j];
		}
	}

	// calc final midpoint
	s->mids[0] = s->mins[0] + (s->maxs[0] - s->mins[0]) * 0.5f;
	s->mids[1] = s->mins[1] + (s->maxs[1] - s->mins[1]) * 0.5f;
	s->mids[2] = s->mins[2] + (s->maxs[2] - s->mins[2]) * 0.5f;
}


void Mod_CalcSurfaceExtents (msurface_t *s, dbsp_t *bsp)
{
	float	mins[2], maxs[2], val;
	int		i, j, e;
	float	*v;
	mtexinfo_t	*tex;
	int		bmins[2], bmaxs[2];

	mins[0] = mins[1] = 999999;
	maxs[0] = maxs[1] = -999999;

	tex = s->texinfo;

	for (i = 0; i < s->numedges; i++)
	{
		e = bsp->surfedges[s->firstedge + i];

		if (e >= 0)
			v = bsp->vertexes[bsp->edges[e].v[0]].point;
		else v = bsp->vertexes[bsp->edges[-e].v[1]].point;

		for (j = 0; j < 2; j++)
		{
			val = Vector3Dot (v, tex->vecs[j]) + tex->vecs[j][3];

			if (val < mins[j]) mins[j] = val;
			if (val > maxs[j]) maxs[j] = val;
		}
	}

	for (i = 0; i < 2; i++)
	{
		bmins[i] = floor (mins[i] / 16);
		bmaxs[i] = ceil (maxs[i] / 16);

		s->texturemins[i] = bmins[i] * 16;
		s->extents[i] = (bmaxs[i] - bmins[i]) * 16;
	}
}


void R_CreateSurfaceLightmap (msurface_t *surf);
void R_EndBuildingLightmaps (void);
void R_BeginBuildingLightmaps (model_t *m);

/*
=================
Mod_LoadFaces
=================
*/
void Mod_LoadFaces (lump_t *l, byte *base, dbsp_t *bsp)
{
	dface_t		*in;
	msurface_t 	*out;
	int			i, count, surfnum;
	int			planenum, side;
	int			ti;

	in = (void *) base;

	if (l->filelen % sizeof (*in))
		ri->Sys_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*in);
	out = ri->Hunk_Alloc (count * sizeof (*out));

	loadmodel->surfaces = out;
	loadmodel->numsurfaces = count;

	R_BeginBuildingSurfaces ();
	R_BeginBuildingLightmaps (loadmodel);

	for (surfnum = 0; surfnum < count; surfnum++, in++, out++)
	{
		out->firstedge = LittleLong (in->firstedge);
		out->numedges = LittleShort (in->numedges);
		out->flags = 0;

		planenum = LittleShort (in->planenum);
		side = LittleShort (in->side);

		if (side)
			out->flags |= SURF_PLANEBACK;

		out->plane = loadmodel->planes + planenum;

		ti = LittleShort (in->texinfo);

		if (ti < 0 || ti >= loadmodel->numtexinfo)
			ri->Sys_Error (ERR_DROP, "MOD_LoadBmodel: bad texinfo number");

		out->texinfo = loadmodel->texinfo + ti;

		Mod_CalcSurfaceExtents (out, bsp);
		Mod_CalcSurfaceBounds (out, bsp);

		// lighting info
		for (i = 0; i < MAXLIGHTMAPS; i++)
			out->styles[i] = in->styles[i];

		i = LittleLong (in->lightofs);

		if (i == -1)
			out->samples = NULL;
		else out->samples = loadmodel->lightdata + i;

		R_RegisterSurface (out);

		// set the drawing flags
		if (out->texinfo->flags & SURF_WARP)
		{
			for (i = 0; i < 2; i++)
			{
				out->extents[i] = 16384;
				out->texturemins[i] = -8192;
			}
		}

		// create lightmaps and polygons
		if (!(out->texinfo->flags & (SURF_SKY | SURF_TRANS33 | SURF_TRANS66 | SURF_WARP)))
			R_CreateSurfaceLightmap (out);
	}

	R_EndBuildingLightmaps ();
	R_EndBuildingSurfaces (loadmodel, bsp);
}


/*
=================
Mod_SetParent
=================
*/
void Mod_SetParent (mnode_t *node, mnode_t *parent)
{
	node->parent = parent;

	if (node->contents != -1)
		return;

	Mod_SetParent (node->children[0], node);
	Mod_SetParent (node->children[1], node);
}

/*
=================
Mod_LoadNodes
=================
*/
void Mod_LoadNodes (lump_t *l, byte *base)
{
	int			i, j, count, p;
	dnode_t		*in;
	mnode_t 	*out;

	in = (void *) base;

	if (l->filelen % sizeof (*in))
		ri->Sys_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*in);
	out = ri->Hunk_Alloc (count * sizeof (*out));

	loadmodel->nodes = out;
	loadmodel->numnodes = count;

	for (i = 0; i < count; i++, in++, out++)
	{
		for (j = 0; j < 3; j++)
		{
			out->mins[j] = LittleShort (in->mins[j]);
			out->maxs[j] = LittleShort (in->maxs[j]);
		}

		p = LittleLong (in->planenum);
		out->plane = loadmodel->planes + p;

		out->firstsurface = LittleShort (in->firstface);
		out->numsurfaces = LittleShort (in->numfaces);
		out->contents = -1;	// differentiate from leafs

		// copy over surfaces for the dlight push
		out->surfaces = loadmodel->surfaces + out->firstsurface;

		// set up child nodes and leafs
		for (j = 0; j < 2; j++)
		{
			p = LittleLong (in->children[j]);

			if (p >= 0)
				out->children[j] = loadmodel->nodes + p;
			else out->children[j] = (mnode_t *) (loadmodel->leafs + (-1 - p));
		}
	}

	Mod_SetParent (loadmodel->nodes, NULL);	// sets nodes and leafs
}


/*
=================
Mod_LoadLeafs
=================
*/
void Mod_LoadLeafs (lump_t *l, byte *base)
{
	dleaf_t 	*in;
	mleaf_t 	*out;
	int			i, j, count, p;

	in = (void *) base;

	if (l->filelen % sizeof (*in))
		ri->Sys_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*in);
	out = ri->Hunk_Alloc (count * sizeof (*out));

	loadmodel->leafs = out;
	loadmodel->numleafs = count;

	for (i = 0; i < count; i++, in++, out++)
	{
		for (j = 0; j < 3; j++)
		{
			out->mins[j] = LittleShort (in->mins[j]);
			out->maxs[j] = LittleShort (in->maxs[j]);
		}

		p = LittleLong (in->contents);
		out->contents = p;

		out->cluster = LittleShort (in->cluster);
		out->area = LittleShort (in->area);

		out->firstmarksurface = loadmodel->marksurfaces + LittleShort (in->firstleafface);
		out->nummarksurfaces = LittleShort (in->numleaffaces);
	}
}


/*
=================
Mod_LoadMarksurfaces
=================
*/
void Mod_LoadMarksurfaces (lump_t *l, byte *base)
{
	int		i, j, count;
	short		*in;
	msurface_t **out;

	in = (void *) base;

	if (l->filelen % sizeof (*in))
		ri->Sys_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*in);
	out = ri->Hunk_Alloc (count * sizeof (*out));

	loadmodel->marksurfaces = out;
	loadmodel->nummarksurfaces = count;

	for (i = 0; i < count; i++)
	{
		j = LittleShort (in[i]);

		if (j < 0 || j >= loadmodel->numsurfaces)
			ri->Sys_Error (ERR_DROP, "Mod_ParseMarksurfaces: bad surface number");

		out[i] = loadmodel->surfaces + j;
	}
}


/*
=================
Mod_LoadPlanes
=================
*/
void Mod_LoadPlanes (lump_t *l, byte *base)
{
	int			i, j;
	cplane_t	*out;
	dplane_t 	*in;
	int			count;
	int			bits;

	in = (void *) base;

	if (l->filelen % sizeof (*in))
		ri->Sys_Error (ERR_DROP, "MOD_LoadBmodel: funny lump size in %s", loadmodel->name);

	count = l->filelen / sizeof (*in);
	out = ri->Hunk_Alloc (count * sizeof (*out));

	loadmodel->planes = out;
	loadmodel->numplanes = count;

	for (i = 0; i < count; i++, in++, out++)
	{
		bits = 0;

		for (j = 0; j < 3; j++)
		{
			out->normal[j] = LittleFloat (in->normal[j]);

			if (out->normal[j] < 0)
				bits |= 1 << j;
		}

		out->dist = LittleFloat (in->dist);
		out->type = LittleLong (in->type);
		out->signbits = bits;
	}
}


/*
=================
Mod_LoadBrushModel
=================
*/
void Mod_LoadBrushModel (model_t *mod, void *buffer)
{
	int			i;
	dbspheader_t *header;
	mmodel_t 	*bm;
	byte		*mod_base;
	dbsp_t		bsp;

	loadmodel->type = mod_brush;

	if (loadmodel != mod_known)
		ri->Sys_Error (ERR_DROP, "Loaded a brush model after the world");

	header = (dbspheader_t *) buffer;

	i = LittleLong (header->version);

	if (i != BSPVERSION)
		ri->Sys_Error (ERR_DROP, "Mod_LoadBrushModel: %s has wrong version number (%i should be %i)", mod->name, i, BSPVERSION);

	// swap all the lumps
	mod_base = (byte *) header;

	for (i = 0; i < sizeof (dbspheader_t) / 4; i++)
		((int *) header)[i] = LittleLong (((int *) header)[i]);

	// load into heap
#define LUMP_BASE(lump) (mod_base + header->lumps[lump].fileofs)
#define DEFINE_LUMP(lump)	&header->lumps[lump], LUMP_BASE (lump)

	// these are only needed at load-time so don't copy them into hunk memory
	bsp.vertexes = (dvertex_t *) LUMP_BASE (LUMP_VERTEXES);
	bsp.edges = (dedge_t *) LUMP_BASE (LUMP_EDGES);
	bsp.surfedges = (int *) LUMP_BASE (LUMP_SURFEDGES);

	// these are needed at runtime so do copy them to hunk memory
	Mod_LoadLighting (DEFINE_LUMP (LUMP_LIGHTING));
	Mod_LoadPlanes (DEFINE_LUMP (LUMP_PLANES));
	Mod_LoadTexinfo (DEFINE_LUMP (LUMP_TEXINFO));
	Mod_LoadFaces (DEFINE_LUMP (LUMP_FACES), &bsp);
	Mod_LoadMarksurfaces (DEFINE_LUMP (LUMP_LEAFFACES));
	Mod_LoadVisibility (DEFINE_LUMP (LUMP_VISIBILITY));
	Mod_LoadLeafs (DEFINE_LUMP (LUMP_LEAFS));
	Mod_LoadNodes (DEFINE_LUMP (LUMP_NODES));
	Mod_LoadSubmodels (DEFINE_LUMP (LUMP_MODELS));

	// regular and alternate animation
	mod->numframes = 2;

	// set up the submodels
	for (i = 0; i < mod->numsubmodels; i++)
	{
		model_t	*starmod;

		bm = &mod->submodels[i];
		starmod = &mod_inline[i];

		// duplicate the basic data
		*starmod = *loadmodel;

		// setup what's different
		starmod->firstmodelsurface = bm->firstface;
		starmod->nummodelsurfaces = bm->numfaces;
		starmod->firstnode = bm->headnode;

		if (starmod->firstnode >= loadmodel->numnodes)
			ri->Sys_Error (ERR_DROP, "Inline model %i has bad firstnode %i", i, starmod->firstnode);

		// this happens in boss1; it would be nice to patch it, but what's the best way???
		// add a dummy node pointing to nothing at the end of the nodes list and use that???
		//if (starmod->firstnode < 0)
		//	ri->Sys_Error (ERR_DROP, "Inline model %i has bad firstnode %i", i, starmod->firstnode);

		VectorCopy (bm->maxs, starmod->maxs);
		VectorCopy (bm->mins, starmod->mins);
		starmod->radius = bm->radius;

		if (i == 0)
			*loadmodel = *starmod;

		starmod->numleafs = bm->visleafs;
	}
}

