/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// C++ layer for D3D object lifecycle management
#include <d3d9.h>
#include <D3Dcompiler.h>
#include <vector>

// needed for the shader compiler
// pD3DCompile is defined in D3Dcompiler.h so we don't need to typedef it ourselves
HINSTANCE hInstCompiler = NULL;
pD3DCompile QD3DCompile = NULL;

static std::vector<IDirect3DVertexShader9 *> d3d_VertexShaders;
static std::vector<IDirect3DPixelShader9 *> d3d_PixelShaders;
static std::vector<IDirect3DVertexDeclaration9 *> d3d_VertexDeclarations;

// so that we can link to C for standard stuff
extern "C" void Com_Printf (char *fmt, ...);
extern "C" void Sys_Error (char *error, ...);
extern "C" char	*va (char *format, ...);

// main D3D objects
extern "C" IDirect3DDevice9 *d3d_Device;

#include "gl_cpplayer.h"
#include "resource.h"

#define STR_EXPAND(tok) #tok
#define DEFINE_SHADER_REGISTER(n) {#n, "c"STR_EXPAND (n)}


static int LoadResourceData (int resourceid, void **resbuf)
{
	// per MSDN, UnlockResource is obsolete and does nothing any more.  There is
	// no way to free the memory used by a resource after you're finished with it.
	// If you ask me this is kinda fucked, but what do I know?  We'll just leak it.
	if (resbuf)
	{
		HRSRC hResInfo = FindResource (NULL, MAKEINTRESOURCE (resourceid), RT_RCDATA);

		if (hResInfo)
		{
			HGLOBAL hResData = LoadResource (NULL, hResInfo);

			if (hResData)
			{
				resbuf[0] = (byte *) LockResource (hResData);
				return SizeofResource (NULL, hResInfo);
			}
		}
	}

	return 0;
}

struct includetable_t
{
	char *name;
	int resourceid;
};

// map include file names to resource ids because there appears to be no way to get a resource by using it's original name
static includetable_t includetable[] = {
	{"fxVSIncludes.fx", IDR_VSINCLUDES},
	{"fxPSIncludes.fx", IDR_PSINCLUDES},
	{"fxCommonIncludes.fx", IDR_COMMONINCLUDES},
	{NULL, 0}
};

class QShaderIncludes : public ID3DInclude
{
public:
	HRESULT WINAPI Open (D3D_INCLUDE_TYPE IncludeType, LPCSTR pFileName, LPCVOID pParentData, LPCVOID *ppData, UINT *pBytes)
	{
		char *included = NULL;
		int includelen = 0;

		// check for specific files we wish to include (file names aren't stored in the exe - boo)
		for (int i = 0;; i++)
		{
			if (!includetable[i].name || !includetable[i].resourceid) break;
			if (strcmp (pFileName, includetable[i].name)) continue;

			includelen = LoadResourceData (includetable[i].resourceid, (void **) &included);
			break;
		}

		if (includelen && included)
		{
			*ppData = included;
			*pBytes = includelen;
		}
		else
		{
			*ppData = NULL;
			*pBytes = 0;
		}

		return S_OK;
	}

	HRESULT WINAPI Close (LPCVOID pData)
	{
		return S_OK;
	}
};


static ID3DBlob *R_CompileShaderCommon (int resourceID, const D3D_SHADER_MACRO *defines, const char *entrypoint, const char *profile)
{
	char *shadertext = NULL;
	ID3DBlob *ShaderBlob = NULL;
	ID3DBlob *ErrorBlob = NULL;
	int textlength = LoadResourceData (resourceID, (void **) &shadertext);

	QShaderIncludes Includes;

	// statically linking to d3dcompiler.lib causes it to use d3dcompiler_47.dll which is not on Windows 7 so link dynamically instead
	if (!hInstCompiler)
	{
		for (int i = 99; i > 32; i--)
		{
			if ((hInstCompiler = LoadLibrary (va ("d3dcompiler_%i.dll", i))) != NULL)
			{
				if ((QD3DCompile = (pD3DCompile) GetProcAddress (hInstCompiler, "D3DCompile")) != NULL)
					break;
				else
				{
					FreeLibrary (hInstCompiler);
					hInstCompiler = NULL;
				}
			}
		}
	}

	if (!hInstCompiler)
	{
		// totally failed to load
		return NULL;
	}

	if (!QD3DCompile)
	{
		// this should never happen
		return NULL;
	}

	HRESULT hr = QD3DCompile (
		shadertext,
		textlength,
		NULL,
		defines,
		&Includes,
		entrypoint,
		profile,
		0,
		0,
		&ShaderBlob,
		&ErrorBlob);

	if (ErrorBlob)
	{
		Com_Printf ("Shader %s error:\n%s\n", entrypoint, (char *) ErrorBlob->GetBufferPointer ());
		ErrorBlob->Release ();
	}
	else if (FAILED (hr))
		Com_Printf ("Terminal shader %s error!!!\n", entrypoint);

	// return the shader blob we got (it can be NULL and the caller will check for it)
	return ShaderBlob;
}


extern "C" IDirect3DVertexShader9 *R_CreateVertexShader (int resourceID, const char *entrypoint)
{
	IDirect3DVertexShader9 *shaderOut = NULL;
	ID3DBlob *ShaderBlob = NULL;

	const D3D_SHADER_MACRO defines[] = {
		{"VERTEXSHADER", "1"},
		{"MAX_LIGHTSTYLES", "256"},			// !!! if these are changed they must be changed in q_shared.h too !!!
		{"MAX_PACKED_LIGHTSTYLES", "64"},	// !!! if these are changed they must be changed in q_shared.h too !!!
		DEFINE_SHADER_REGISTER (VSC_WORLDMATRIX),
		DEFINE_SHADER_REGISTER (VSC_ENTITYMATRIX),
		DEFINE_SHADER_REGISTER (VSC_LOCALMATRIX),
		DEFINE_SHADER_REGISTER (VSC_WARPTIME),
		DEFINE_SHADER_REGISTER (VSC_VUP),
		DEFINE_SHADER_REGISTER (VSC_VPN),
		DEFINE_SHADER_REGISTER (VSC_VRIGHT),
		DEFINE_SHADER_REGISTER (VSC_RORIGIN),
		DEFINE_SHADER_REGISTER (VSC_SPRITEORIGIN),
		DEFINE_SHADER_REGISTER (VSC_SKYMATRIX),
		DEFINE_SHADER_REGISTER (VSC_TEXSCROLL),
		DEFINE_SHADER_REGISTER (VSC_TRANSFORMEDDL),
		DEFINE_SHADER_REGISTER (VSC_SURFACEPASS),
		DEFINE_SHADER_REGISTER (VSC_ORTHOMATRIX),
		DEFINE_SHADER_REGISTER (VSC_RTINVERSE),
		DEFINE_SHADER_REGISTER (VSC_MESHMOVE),
		DEFINE_SHADER_REGISTER (VSC_MESHFRONTV),
		DEFINE_SHADER_REGISTER (VSC_MESHBACKV),
		DEFINE_SHADER_REGISTER (VSC_MESHSUITSCALE),
		DEFINE_SHADER_REGISTER (VSC_MESHBLEND),
		DEFINE_SHADER_REGISTER (VSC_LIGHTSTYLES),
		{NULL, NULL}
	};

	if ((ShaderBlob = R_CompileShaderCommon (resourceID, defines, entrypoint, "vs_3_0")) != NULL)
	{
		if (SUCCEEDED (d3d_Device->CreateVertexShader ((DWORD *) ShaderBlob->GetBufferPointer (), &shaderOut)))
			d3d_VertexShaders.push_back (shaderOut);

		ShaderBlob->Release ();
	}

	return shaderOut;
}


extern "C" IDirect3DPixelShader9 *R_CreatePixelShader (int resourceID, const char *entrypoint)
{
	IDirect3DPixelShader9 *shaderOut = NULL;
	ID3DBlob *ShaderBlob = NULL;

	const D3D_SHADER_MACRO defines[] = {
		{"PIXELSHADER", "1"},
		DEFINE_SHADER_REGISTER (PSC_POLYBLEND),
		DEFINE_SHADER_REGISTER (PSC_SHADELIGHT),
		DEFINE_SHADER_REGISTER (PSC_SHADEVECTOR),
		DEFINE_SHADER_REGISTER (PSC_LIGHTCOLOR),
		DEFINE_SHADER_REGISTER (PSC_LIGHTRADIUS),
		DEFINE_SHADER_REGISTER (PSC_ALPHAVAL),
		DEFINE_SHADER_REGISTER (PSC_WARPTIME),
		DEFINE_SHADER_REGISTER (PSC_GAMMA),
		DEFINE_SHADER_REGISTER (PSC_INTENSITY),
		DEFINE_SHADER_REGISTER (PSC_BRIGHTNESS),
		{NULL, NULL}
	};

	if ((ShaderBlob = R_CompileShaderCommon (resourceID, defines, entrypoint, "ps_3_0")) != NULL)
	{
		if (SUCCEEDED (d3d_Device->CreatePixelShader ((DWORD *) ShaderBlob->GetBufferPointer (), &shaderOut)))
			d3d_PixelShaders.push_back (shaderOut);

		ShaderBlob->Release ();
	}

	return shaderOut;
}


extern "C" IDirect3DVertexDeclaration9 *R_CreateVertexDeclaration (D3DVERTEXELEMENT9 *layout)
{
	IDirect3DVertexDeclaration9 *vd = NULL;

	if (SUCCEEDED (d3d_Device->CreateVertexDeclaration (layout, &vd)))
		d3d_VertexDeclarations.push_back (vd);

	return vd;
}


extern "C" void R_ReleaseShaders (void)
{
	// so i never have to worry about releasing stuff
	for (auto vd : d3d_VertexDeclarations) SAFE_RELEASE (vd);
	for (auto vs : d3d_VertexShaders) SAFE_RELEASE (vs);
	for (auto ps : d3d_PixelShaders) SAFE_RELEASE (ps);

	// and clear the vectors
	d3d_VertexShaders.clear ();
	d3d_PixelShaders.clear ();
	d3d_VertexDeclarations.clear ();

	// any future calls into here are errors
	QD3DCompile = NULL;

	// and unload the library
	if (hInstCompiler)
	{
		FreeLibrary (hInstCompiler);
		hInstCompiler = NULL;
	}
}


extern "C" void R_SetVertexShaderConstant4fv (int reg, float *c, int num4f)
{
	d3d_Device->SetVertexShaderConstantF (reg, c, num4f);
}

extern "C" void R_SetVertexShaderConstant4f (int reg, float c0, float c1, float c2, float c3)
{
	float c[] = {c0, c1, c2, c3};
	d3d_Device->SetVertexShaderConstantF (reg, c, 1);
}

extern "C" void R_SetVertexShaderConstant3f (int reg, float c0, float c1, float c2)
{
	float c[] = {c0, c1, c2, 0};
	d3d_Device->SetVertexShaderConstantF (reg, c, 1);
}

extern "C" void R_SetVertexShaderConstant2f (int reg, float c0, float c1)
{
	float c[] = {c0, c1, 0, 0};
	d3d_Device->SetVertexShaderConstantF (reg, c, 1);
}

extern "C" void R_SetVertexShaderConstant1f (int reg, float c0)
{
	float c[] = {c0, 0, 0, 0};
	d3d_Device->SetVertexShaderConstantF (reg, c, 1);
}

extern "C" void R_SetPixelShaderConstant4fv (int reg, float *c, int num4f)
{
	d3d_Device->SetPixelShaderConstantF (reg, c, num4f);
}

extern "C" void R_SetPixelShaderConstant4f (int reg, float c0, float c1, float c2, float c3)
{
	float c[] = {c0, c1, c2, c3};
	d3d_Device->SetPixelShaderConstantF (reg, c, 1);
}

extern "C" void R_SetPixelShaderConstant3f (int reg, float c0, float c1, float c2)
{
	float c[] = {c0, c1, c2, 0};
	d3d_Device->SetPixelShaderConstantF (reg, c, 1);
}

extern "C" void R_SetPixelShaderConstant2f (int reg, float c0, float c1)
{
	float c[] = {c0, c1, 0, 0};
	d3d_Device->SetPixelShaderConstantF (reg, c, 1);
}

extern "C" void R_SetPixelShaderConstant1f (int reg, float c0)
{
	float c[] = {c0, 0, 0, 0};
	d3d_Device->SetPixelShaderConstantF (reg, c, 1);
}


