/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "gl_local.h"

// to do: load mip levels from WAL files

void *R_Malloc (int size)
{
	// this just wraps ri->Scratch_AllocSize so that i can pass a custom allocator to LoadPCX.  Yuck.
	return ri->Scratch_AllocSize (size);
}


image_t		gltextures[MAX_GLTEXTURES];
int			numgltextures;

unsigned	d_8to24table[256];


typedef struct glmode_s
{
	char *name;
	D3DTEXTUREFILTERTYPE minmagfilter;
	D3DTEXTUREFILTERTYPE mipfilter;
} glmode_t;

glmode_t modes[] =
{
	{"GL_NEAREST", D3DTEXF_POINT, D3DTEXF_NONE},
	{"GL_LINEAR", D3DTEXF_LINEAR, D3DTEXF_NONE},
	{"GL_NEAREST_MIPMAP_NEAREST", D3DTEXF_POINT, D3DTEXF_POINT},
	{"GL_LINEAR_MIPMAP_NEAREST", D3DTEXF_LINEAR, D3DTEXF_POINT},
	{"GL_NEAREST_MIPMAP_LINEAR", D3DTEXF_POINT, D3DTEXF_LINEAR},
	{"GL_LINEAR_MIPMAP_LINEAR", D3DTEXF_LINEAR, D3DTEXF_LINEAR}
};

#define NUM_GL_MODES (sizeof (modes) / sizeof (glmode_t))


void R_SetTextureMode (glmode_t *mode, int anisotropy)
{
	// update diffuse and sky samplers
	R_SetSamplerState (TS_DIFFUSE, mode->minmagfilter, mode->mipfilter, D3DTADDRESS_WRAP, anisotropy);
	R_SetSamplerState (TS_SKY, mode->minmagfilter, D3DTEXF_NONE, D3DTADDRESS_CLAMP, 1);

	// store back to the cvar and flag as unmodified so it won't fire again next frame
	ri->Cvar_Set ("gl_texturemode", mode->name);
	gl_texturemode->modified = false;
}


/*
===============
R_TextureMode
===============
*/
void R_TextureMode (char *string, int anisotropy)
{
	if (!string || !string[0]) return;

	switch (string[0])
	{
	case '0': R_SetTextureMode (&modes[0], anisotropy); return;
	case '1': R_SetTextureMode (&modes[1], anisotropy); return;
	case '2': R_SetTextureMode (&modes[2], anisotropy); return;
	case '3': R_SetTextureMode (&modes[3], anisotropy); return;
	case '4': R_SetTextureMode (&modes[4], anisotropy); return;
	case '5': R_SetTextureMode (&modes[5], anisotropy); return;
	default: break;
	}

	for (int i = 0; i < NUM_GL_MODES; i++)
	{
		if (!Q_stricmp (modes[i].name, string))
		{
			R_SetTextureMode (&modes[i], anisotropy);
			return;
		}
	}

	ri->Con_Printf (PRINT_ALL, "bad filter name: \"%s\"\n", string);
}


/*
===============
R_ImageList_f
===============
*/
void R_ImageList_f (void)
{
	int		i;
	image_t	*image;
	int		texels;

	ri->Con_Printf (PRINT_ALL, "------------------\n");
	texels = 0;

	for (i = 0, image = gltextures; i < numgltextures; i++, image++)
	{
		D3DSURFACE_DESC desc;

		if (!image->texture) continue;

		if (SUCCEEDED (image->texture->lpVtbl->GetLevelDesc (image->texture, 0, &desc)))
		{
			int upload_width = desc.Width;
			int upload_height = desc.Height;

			texels += upload_width * upload_height;

			switch (image->type)
			{
			case it_skin:	ri->Con_Printf (PRINT_ALL, "M"); break;
			case it_sprite: ri->Con_Printf (PRINT_ALL, "S"); break;
			case it_wall:	ri->Con_Printf (PRINT_ALL, "W"); break;
			case it_pic:	ri->Con_Printf (PRINT_ALL, "P"); break;
			default:		ri->Con_Printf (PRINT_ALL, " "); break;
			}

			ri->Con_Printf (PRINT_ALL, " %3i %3i: %s\n", upload_width, upload_height, image->name);
		}
	}

	ri->Con_Printf (PRINT_ALL, "Total texel count (not counting mipmaps): %i\n", texels);
}


image_t *R_GetFreeImage (char *name, int width, int height, imagetype_t type)
{
	int i;

	// find a free image_t
	for (i = 0; i < numgltextures; i++)
	{
		if (!gltextures[i].texture)
			break;
	}

	if (i == numgltextures)
	{
		if (numgltextures == MAX_GLTEXTURES)
			ri->Sys_Error (ERR_DROP, "MAX_GLTEXTURES");

		numgltextures++;
	}

	image_t *image = &gltextures[i];

	if (strlen (name) >= sizeof (image->name))
		ri->Sys_Error (ERR_DROP, "Draw_LoadPic: \"%s\" is too long", name);

	strcpy (image->name, name);
	image->registration_sequence = r_registration_sequence;

	image->width = width;
	image->height = height;
	image->type = type;
	image->mipmap = (image->type != it_pic && image->type != it_sky);

	return image;
}


/*
================
R_LoadPic

This is also used as an entry point for the generated r_notexture
================
*/
image_t *R_LoadPic (char *name, byte *pic, int width, int height, imagetype_t type, unsigned *palette)
{
	image_t *image = R_GetFreeImage (name, width, height, type);

	if (image->type == it_skin && palette)
		R_FloodFillSkin (pic, width, height, palette);

	qboolean alpha = (image->type == it_pic || image->type == it_sprite);

	if ((image->texture = R_CreateTexture (width, height, image->mipmap, alpha)) != NULL)
		R_FillTexture (image->texture, pic, width, height, palette, alpha);

	return image;
}


image_t *R_SearchImages (char *name, imagetype_t type)
{
	for (int i = 0; i < numgltextures; i++)
	{
		image_t *image = &gltextures[i];

		// not a match
		if (!image->texture) continue;
		if (image->type != type) continue;
		if (_stricmp (image->name, name)) continue;

		// matched
		return image;
	}

	// no match found
	return NULL;
}


/*
===============
R_FindImage

Finds or loads the given image
===============
*/
image_t *R_FindImage (char *name, imagetype_t type)
{
	image_t	*image;
	int		len;
	int		width, height;
	int		mark = ri->Scratch_LowMark ();

	if (!name) return NULL;
	if ((len = strlen (name)) < 5) return NULL;

	// look for it
	if ((image = R_SearchImages (name, type)) != NULL)
	{
		image->registration_sequence = r_registration_sequence;
		return image;
	}

	// load the pic from disk
	byte *pic = NULL;
	byte *palette = NULL;
	char *ext = name + len - 4;

	if (!strcmp (ext, ".pcx"))
	{
		R_LoadPCX (name, &pic, &palette, &width, &height, R_Malloc);

		if (!pic)
			image = r_notexture;
		else image = R_LoadPic (name, pic, width, height, type, R_Palette24To32 (palette));
	}
	else if (!strcmp (ext, ".wal"))
	{
		R_LoadWAL (name, &pic, &width, &height, R_Malloc);

		if (!pic)
			image = r_notexture;
		else image = R_LoadPic (name, pic, width, height, type, d_8to24table);
	}
	else if (!strcmp (ext, ".tga"))
	{
		R_LoadTGA (name, &pic, &width, &height, R_Malloc);

		if (!pic)
			image = r_notexture;
		else image = R_LoadPic (name, pic, width, height, type, NULL);
	}
	else image = r_notexture;

	ri->Scratch_FreeToLowMark (mark);

	return image;
}


/*
===============
R_RegisterSkin
===============
*/
struct image_s *R_RegisterSkin (char *name)
{
	return R_FindImage (name, it_skin);
}


void R_ReleaseTexture (image_t *image)
{
	SAFE_RELEASE (image->texture);
	memset (image, 0, sizeof (*image));
}


/*
================
R_FreeUnusedImages

Any image that was not touched on this registration sequence
will be freed.
================
*/
void R_FreeUnusedImages (void)
{
	// never free these textures
	r_notexture->registration_sequence = r_registration_sequence;
	r_blacktexture->registration_sequence = r_registration_sequence;
	r_greytexture->registration_sequence = r_registration_sequence;
	r_whitetexture->registration_sequence = r_registration_sequence;

	for (int i = 0; i < numgltextures; i++)
	{
		image_t *image = &gltextures[i];

		if (image->registration_sequence == r_registration_sequence) continue;	// used this sequence
		if (!image->registration_sequence) continue;							// free image_t slot
		if (!image->texture) continue;											// free image_t slot
		if (image->type == it_pic) continue;									// don't free pics

		R_ReleaseTexture (image);
	}
}


/*
===============
Draw_GetPalette

===============
*/
int Draw_GetPalette (void)
{
	int	width, height;
	byte *pic = NULL, *pal = NULL;
	int mark = ri->Scratch_LowMark ();

	// get the palette
	R_LoadPCX ("pics/colormap.pcx", &pic, &pal, &width, &height, R_Malloc);

	// validate the colormap
	if (!pic) { ri->Sys_Error (ERR_FATAL, "Couldn't load pics/colormap.pcx"); return 0; }
	if (!pal) { ri->Sys_Error (ERR_FATAL, "Couldn't load pics/colormap.pcx"); return 0; }
	if (width != 256) { ri->Sys_Error (ERR_FATAL, "Bad colormap"); return 0; }
	if (height < 320) { ri->Sys_Error (ERR_FATAL, "Bad colormap"); return 0; }

	// convert the palette up
	memcpy (d_8to24table, R_Palette24To32 (pal), sizeof (d_8to24table));

	ri->Scratch_FreeToLowMark (mark);

	return 0;
}


/*
===============
R_InitImages
===============
*/
void R_InitImages (void)
{
	// go to a new registration sequence
	r_registration_sequence++;

	Draw_GetPalette ();

	// this was never reset on a vid_restart
	numgltextures = 0;
}


/*
===============
R_ShutdownImages
===============
*/
void R_ShutdownImages (void)
{
	for (int i = 0; i < MAX_GLTEXTURES; i++)
		R_ReleaseTexture (&gltextures[i]);

	numgltextures = 0;
}


char *R_NameForImage (struct image_s *image)
{
	// this only exists so the checks in CL_AddPacketEntities can be less hacky
	return image->name;
}

