/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// r_main.c

#include "gl_local.h"
#include "resource.h"


IDirect3DVertexDeclaration9 *d3d_PartDecl = NULL;

IDirect3DVertexShader9 *d3d_PartVS = NULL;
IDirect3DPixelShader9 *d3d_PartPS = NULL;

IDirect3DVertexBuffer9 *d3d_PartInstanceBuffer = NULL;
IDirect3DVertexBuffer9 *d3d_PartVertexBuffer = NULL;
IDirect3DIndexBuffer9 *d3d_PartIndexBuffer = NULL;

int r_firstparticle = 0;

void Part_OnCreate (void)
{
	uint16_t partindexes[] = {0, 1, 2, 3};
	float partpervert[] = {-1, -1, -1, 1, 1, -1, 1, 1};
	void *data = NULL;

	D3DVERTEXELEMENT9 layout[] = {
		VDECL (1, offsetof (particle_t, origin), D3DDECLTYPE_FLOAT3, D3DDECLUSAGE_POSITION, 0),
		VDECL (1, offsetof (particle_t, color), D3DDECLTYPE_D3DCOLOR, D3DDECLUSAGE_COLOR, 0),
		VDECL (1, offsetof (particle_t, alpha), D3DDECLTYPE_FLOAT1, D3DDECLUSAGE_TEXCOORD, 1),
		VDECL (0, 0, D3DDECLTYPE_FLOAT2, D3DDECLUSAGE_TEXCOORD, 0),
		D3DDECL_END ()
	};

	d3d_PartDecl = R_CreateVertexDeclaration (layout);

	d3d_Device->lpVtbl->CreateVertexBuffer (d3d_Device, sizeof (partpervert), D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &d3d_PartVertexBuffer, NULL);
	d3d_Device->lpVtbl->CreateIndexBuffer (d3d_Device, sizeof (partindexes), D3DUSAGE_WRITEONLY, D3DFMT_INDEX16, D3DPOOL_MANAGED, &d3d_PartIndexBuffer, NULL);

	if (SUCCEEDED (d3d_PartVertexBuffer->lpVtbl->Lock (d3d_PartVertexBuffer, 0, 0, &data, 0)))
	{
		memcpy (data, partpervert, sizeof (partpervert));
		d3d_PartVertexBuffer->lpVtbl->Unlock (d3d_PartVertexBuffer);
	}

	if (SUCCEEDED (d3d_PartIndexBuffer->lpVtbl->Lock (d3d_PartIndexBuffer, 0, 0, &data, 0)))
	{
		memcpy (data, partindexes, sizeof (partindexes));
		d3d_PartIndexBuffer->lpVtbl->Unlock (d3d_PartIndexBuffer);
	}

	R_CreateShaders (PartVS, PartPS, IDR_PARTSHADER);
}


void Part_OnReset (void)
{
	d3d_Device->lpVtbl->CreateVertexBuffer (
		d3d_Device,
		sizeof (particle_t) * MAX_PARTICLES,
		D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY,
		0,
		D3DPOOL_DEFAULT,
		&d3d_PartInstanceBuffer,
		NULL);

	r_firstparticle = 0;
}


void Part_OnLost (void)
{
	SAFE_RELEASE (d3d_PartInstanceBuffer);
}


void Part_OnDestroy (void)
{
	SAFE_RELEASE (d3d_PartVertexBuffer);
	SAFE_RELEASE (d3d_PartIndexBuffer);
}


void Part_Register (void)
{
	static qboolean registered = false;

	if (!registered)
	{
		R_RegisterHandler (Part_OnCreate, Part_OnLost, Part_OnReset, Part_OnDestroy);
		registered = true;
	}
}


/*
===============
R_DrawParticles
===============
*/
void R_DrawParticles (refdef_t *rd)
{
	if (!rd->num_particles) return;

	// find the locking mode
	particle_t *pverts = NULL;
	DWORD vblock = D3DLOCK_NOOVERWRITE | D3DLOCK_NOSYSLOCK;

	if (r_firstparticle + rd->num_particles >= MAX_PARTICLES)
	{
		vblock = D3DLOCK_DISCARD | D3DLOCK_NOSYSLOCK;
		r_firstparticle = 0;
	}

	UINT offset = r_firstparticle * sizeof (particle_t);
	UINT size = rd->num_particles * sizeof (particle_t);

	// lock for the number of particles in the appropriate mode
	if (FAILED (d3d_PartInstanceBuffer->lpVtbl->Lock (d3d_PartInstanceBuffer, offset, size, (void **) &pverts, vblock))) return;

	// copy them in
	memcpy (pverts, rd->particles, size);
	d3d_PartInstanceBuffer->lpVtbl->Unlock (d3d_PartInstanceBuffer);

	// set up state (only doing this if the lock succeeded)
	R_SetRenderStates (RS_DEPTHTEST | RS_CULLCCW | RS_ALPHABLEND | RS_ALPHATEST);

	// shaders
	R_SetVertexDeclaration (d3d_PartDecl);
	R_SetVertexShader (d3d_PartVS);
	R_SetPixelShader (d3d_PartPS);

	// buffers
	// offset into the per-instance buffer for the first particle and set the number of instances to draw
	R_SetVertexBuffer (0, d3d_PartVertexBuffer, 0, sizeof (float) * 2);
	R_SetVertexBuffer (1, d3d_PartInstanceBuffer, offset, sizeof (particle_t));
	R_SetIndexBuffer (d3d_PartIndexBuffer);

	// instancing (the freq for the per-vert buffer depends on the rd->num_particles so it must be set at draw time)
	d3d_Device->lpVtbl->SetStreamSourceFreq (d3d_Device, 0, D3DSTREAMSOURCE_INDEXEDDATA | rd->num_particles);	// num instances
	d3d_Device->lpVtbl->SetStreamSourceFreq (d3d_Device, 1, D3DSTREAMSOURCE_INSTANCEDATA | 1);					// step rate

	d3d_Device->lpVtbl->DrawIndexedPrimitive (d3d_Device, D3DPT_TRIANGLESTRIP, 0, 0, 4, 0, 2);
	c_draw_call++;

	// advance count
	r_firstparticle += rd->num_particles;

	// no more instancing
	d3d_Device->lpVtbl->SetStreamSourceFreq (d3d_Device, 0, 1);
	d3d_Device->lpVtbl->SetStreamSourceFreq (d3d_Device, 1, 1);
}


