/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// draw.c

#include "gl_local.h"


void R_SetViewport (D3DVIEWPORT9 *vp)
{
	// (1) set the viewport
	d3d_Device->lpVtbl->SetViewport (d3d_Device, vp);

	// (2) correct the half-pixel offset: http://aras-p.info/blog/2016/04/08/solving-dx9-half-pixel-offset/
	R_SetVertexShaderConstant2f (VSC_RTINVERSE, 1.0f / vp->Width, 1.0f / vp->Height);
}


void R_SetIndexBuffer (IDirect3DIndexBuffer9 *ib)
{
	if (gl_state.CurrentIB != ib)
	{
		d3d_Device->lpVtbl->SetIndices (d3d_Device, ib);
		gl_state.CurrentIB = ib;
	}
}


typedef struct vertexstream_s
{
	IDirect3DVertexBuffer9 *pStreamData;
	UINT OffsetInBytes;
	UINT Stride;
} vertexstream_t;

#define MAX_VERTEX_STREAMS	16

vertexstream_t d3d_VertexStreams[MAX_VERTEX_STREAMS];

void R_SetVertexBuffer (UINT StreamNumber, IDirect3DVertexBuffer9 *pStreamData, UINT OffsetInBytes, UINT Stride)
{
	if (d3d_VertexStreams[StreamNumber].pStreamData != pStreamData || d3d_VertexStreams[StreamNumber].OffsetInBytes != OffsetInBytes || d3d_VertexStreams[StreamNumber].Stride != Stride)
	{
		d3d_Device->lpVtbl->SetStreamSource (d3d_Device, StreamNumber, pStreamData, OffsetInBytes, Stride);
		d3d_VertexStreams[StreamNumber].pStreamData = pStreamData;
		d3d_VertexStreams[StreamNumber].OffsetInBytes = OffsetInBytes;
		d3d_VertexStreams[StreamNumber].Stride = Stride;
	}
}


void R_SetDPUPMode (void)
{
	// using DrawPrimitiveUP will unbind all streams so we must clear the caches
	for (int i = 0; i < MAX_VERTEX_STREAMS; i++)
		d3d_VertexStreams[i].pStreamData = NULL;

	gl_state.CurrentIB = NULL;
}


DWORD d3d_RenderStates[256];

// !!!! this should never be called outside of R_SetRenderStates !!!!
static void R_SetRenderState (D3DRENDERSTATETYPE rs, DWORD value)
{
	if (d3d_RenderStates[rs] != value)
	{
		d3d_Device->lpVtbl->SetRenderState (d3d_Device, rs, value);
		d3d_RenderStates[rs] = value;
	}
}


void R_SetScissor (BOOL enable, RECT *sr)
{
	// enable or disable scissor test
	R_SetRenderState (D3DRS_SCISSORTESTENABLE, enable);

	if (enable && sr)
	{
		// set the scissor rect if requested
		d3d_Device->lpVtbl->SetScissorRect (d3d_Device, sr);
	}
}


typedef struct stagecache_s
{
	IDirect3DBaseTexture9 *Texture;
} stagecache_t;

stagecache_t d3d_StageCache[8];


static void R_SetBaseTexture (UINT stage, IDirect3DBaseTexture9 *texture)
{
	if (d3d_StageCache[stage].Texture != texture)
	{
		d3d_Device->lpVtbl->SetTexture (d3d_Device, stage, texture);
		d3d_StageCache[stage].Texture = texture;
	}
}


void R_SetTexture (UINT stage, image_t *image)
{
	R_SetBaseTexture (stage, (IDirect3DBaseTexture9 *) image->texture);
}


void R_SetCubeTexture (UINT stage, IDirect3DCubeTexture9 *texture)
{
	R_SetBaseTexture (stage, (IDirect3DBaseTexture9 *) texture);
}


IDirect3DVertexShader9 *d3d_CurrentVS = NULL;
IDirect3DPixelShader9 *d3d_CurrentPS = NULL;


void R_SetVertexShader (IDirect3DVertexShader9 *vs)
{
	if (d3d_CurrentVS != vs)
	{
		d3d_Device->lpVtbl->SetVertexShader (d3d_Device, vs);
		d3d_CurrentVS = vs;
	}
}


void R_SetPixelShader (IDirect3DPixelShader9 *ps)
{
	if (d3d_CurrentPS != ps)
	{
		d3d_Device->lpVtbl->SetPixelShader (d3d_Device, ps);
		d3d_CurrentPS = ps;
	}
}


void R_SetVertexDeclaration (IDirect3DVertexDeclaration9 *decl)
{
	if (gl_state.CurrentDecl != decl)
	{
		d3d_Device->lpVtbl->SetVertexDeclaration (d3d_Device, decl);
		gl_state.CurrentDecl = decl;
	}
}


void R_SetSamplerState (UINT stage, D3DTEXTUREFILTERTYPE minmagfilter, D3DTEXTUREFILTERTYPE mipfilter, D3DTEXTUREADDRESS addressmode, DWORD anisotropy)
{
	int aniso = 1;

	// get anisotropy as a power of 2 and bound to device maximum
	for (aniso = 1; aniso < anisotropy; aniso <<= 1);
	if (aniso > d3d_DeviceCaps.MaxAnisotropy) aniso = d3d_DeviceCaps.MaxAnisotropy;

	if (aniso > 1)
	{
		d3d_Device->lpVtbl->SetSamplerState (d3d_Device, stage, D3DSAMP_MINFILTER, D3DTEXF_ANISOTROPIC);
		d3d_Device->lpVtbl->SetSamplerState (d3d_Device, stage, D3DSAMP_MAGFILTER, D3DTEXF_ANISOTROPIC);
		d3d_Device->lpVtbl->SetSamplerState (d3d_Device, stage, D3DSAMP_MAXANISOTROPY, aniso);
	}
	else
	{
		d3d_Device->lpVtbl->SetSamplerState (d3d_Device, stage, D3DSAMP_MINFILTER, minmagfilter);
		d3d_Device->lpVtbl->SetSamplerState (d3d_Device, stage, D3DSAMP_MAGFILTER, minmagfilter);
		d3d_Device->lpVtbl->SetSamplerState (d3d_Device, stage, D3DSAMP_MAXANISOTROPY, 1);
	}

	d3d_Device->lpVtbl->SetSamplerState (d3d_Device, stage, D3DSAMP_MIPFILTER, mipfilter);

	d3d_Device->lpVtbl->SetSamplerState (d3d_Device, stage, D3DSAMP_ADDRESSU, addressmode);
	d3d_Device->lpVtbl->SetSamplerState (d3d_Device, stage, D3DSAMP_ADDRESSV, addressmode);
	d3d_Device->lpVtbl->SetSamplerState (d3d_Device, stage, D3DSAMP_ADDRESSW, addressmode);
}


/*
===================
R_SetDefaultState
===================
*/
void R_SetDefaultState (void)
{
	// get initial render states
#define GETRENDERSTATE(rs) \
	d3d_Device->lpVtbl->GetRenderState (d3d_Device, rs, &d3d_RenderStates[rs]);

	GETRENDERSTATE (D3DRS_ZENABLE);
	GETRENDERSTATE (D3DRS_FILLMODE);
	GETRENDERSTATE (D3DRS_SHADEMODE);
	GETRENDERSTATE (D3DRS_ZWRITEENABLE);
	GETRENDERSTATE (D3DRS_ALPHATESTENABLE);
	GETRENDERSTATE (D3DRS_LASTPIXEL);
	GETRENDERSTATE (D3DRS_SRCBLEND);
	GETRENDERSTATE (D3DRS_DESTBLEND);
	GETRENDERSTATE (D3DRS_CULLMODE);
	GETRENDERSTATE (D3DRS_ZFUNC);
	GETRENDERSTATE (D3DRS_ALPHAREF);
	GETRENDERSTATE (D3DRS_ALPHAFUNC);
	GETRENDERSTATE (D3DRS_DITHERENABLE);
	GETRENDERSTATE (D3DRS_ALPHABLENDENABLE);
	GETRENDERSTATE (D3DRS_FOGENABLE);
	GETRENDERSTATE (D3DRS_SPECULARENABLE);
	GETRENDERSTATE (D3DRS_FOGCOLOR);
	GETRENDERSTATE (D3DRS_FOGTABLEMODE);
	GETRENDERSTATE (D3DRS_FOGSTART);
	GETRENDERSTATE (D3DRS_FOGEND);
	GETRENDERSTATE (D3DRS_FOGDENSITY);
	GETRENDERSTATE (D3DRS_RANGEFOGENABLE);
	GETRENDERSTATE (D3DRS_STENCILENABLE);
	GETRENDERSTATE (D3DRS_STENCILFAIL);
	GETRENDERSTATE (D3DRS_STENCILZFAIL);
	GETRENDERSTATE (D3DRS_STENCILPASS);
	GETRENDERSTATE (D3DRS_STENCILFUNC);
	GETRENDERSTATE (D3DRS_STENCILREF);
	GETRENDERSTATE (D3DRS_STENCILMASK);
	GETRENDERSTATE (D3DRS_STENCILWRITEMASK);
	GETRENDERSTATE (D3DRS_TEXTUREFACTOR);
	GETRENDERSTATE (D3DRS_WRAP0);
	GETRENDERSTATE (D3DRS_WRAP1);
	GETRENDERSTATE (D3DRS_WRAP2);
	GETRENDERSTATE (D3DRS_WRAP3);
	GETRENDERSTATE (D3DRS_WRAP4);
	GETRENDERSTATE (D3DRS_WRAP5);
	GETRENDERSTATE (D3DRS_WRAP6);
	GETRENDERSTATE (D3DRS_WRAP7);
	GETRENDERSTATE (D3DRS_CLIPPING);
	GETRENDERSTATE (D3DRS_LIGHTING);
	GETRENDERSTATE (D3DRS_AMBIENT);
	GETRENDERSTATE (D3DRS_FOGVERTEXMODE);
	GETRENDERSTATE (D3DRS_COLORVERTEX);
	GETRENDERSTATE (D3DRS_LOCALVIEWER);
	GETRENDERSTATE (D3DRS_NORMALIZENORMALS);
	GETRENDERSTATE (D3DRS_DIFFUSEMATERIALSOURCE);
	GETRENDERSTATE (D3DRS_SPECULARMATERIALSOURCE);
	GETRENDERSTATE (D3DRS_AMBIENTMATERIALSOURCE);
	GETRENDERSTATE (D3DRS_EMISSIVEMATERIALSOURCE);
	GETRENDERSTATE (D3DRS_VERTEXBLEND);
	GETRENDERSTATE (D3DRS_CLIPPLANEENABLE);
	GETRENDERSTATE (D3DRS_POINTSIZE);
	GETRENDERSTATE (D3DRS_POINTSIZE_MIN);
	GETRENDERSTATE (D3DRS_POINTSPRITEENABLE);
	GETRENDERSTATE (D3DRS_POINTSCALEENABLE);
	GETRENDERSTATE (D3DRS_POINTSCALE_A);
	GETRENDERSTATE (D3DRS_POINTSCALE_B);
	GETRENDERSTATE (D3DRS_POINTSCALE_C);
	GETRENDERSTATE (D3DRS_MULTISAMPLEANTIALIAS);
	GETRENDERSTATE (D3DRS_MULTISAMPLEMASK);
	GETRENDERSTATE (D3DRS_PATCHEDGESTYLE);
	GETRENDERSTATE (D3DRS_DEBUGMONITORTOKEN);
	GETRENDERSTATE (D3DRS_POINTSIZE_MAX);
	GETRENDERSTATE (D3DRS_INDEXEDVERTEXBLENDENABLE);
	GETRENDERSTATE (D3DRS_COLORWRITEENABLE);
	GETRENDERSTATE (D3DRS_TWEENFACTOR);
	GETRENDERSTATE (D3DRS_BLENDOP);
	GETRENDERSTATE (D3DRS_POSITIONDEGREE);
	GETRENDERSTATE (D3DRS_NORMALDEGREE);
	GETRENDERSTATE (D3DRS_SCISSORTESTENABLE);
	GETRENDERSTATE (D3DRS_SLOPESCALEDEPTHBIAS);
	GETRENDERSTATE (D3DRS_ANTIALIASEDLINEENABLE);
	GETRENDERSTATE (D3DRS_MINTESSELLATIONLEVEL);
	GETRENDERSTATE (D3DRS_MAXTESSELLATIONLEVEL);
	GETRENDERSTATE (D3DRS_ADAPTIVETESS_X);
	GETRENDERSTATE (D3DRS_ADAPTIVETESS_Y);
	GETRENDERSTATE (D3DRS_ADAPTIVETESS_Z);
	GETRENDERSTATE (D3DRS_ADAPTIVETESS_W);
	GETRENDERSTATE (D3DRS_ENABLEADAPTIVETESSELLATION);
	GETRENDERSTATE (D3DRS_TWOSIDEDSTENCILMODE);
	GETRENDERSTATE (D3DRS_CCW_STENCILFAIL);
	GETRENDERSTATE (D3DRS_CCW_STENCILZFAIL);
	GETRENDERSTATE (D3DRS_CCW_STENCILPASS);
	GETRENDERSTATE (D3DRS_CCW_STENCILFUNC);
	GETRENDERSTATE (D3DRS_COLORWRITEENABLE1);
	GETRENDERSTATE (D3DRS_COLORWRITEENABLE2);
	GETRENDERSTATE (D3DRS_COLORWRITEENABLE3);
	GETRENDERSTATE (D3DRS_BLENDFACTOR);
	GETRENDERSTATE (D3DRS_SRGBWRITEENABLE);
	GETRENDERSTATE (D3DRS_DEPTHBIAS);
	GETRENDERSTATE (D3DRS_WRAP8);
	GETRENDERSTATE (D3DRS_WRAP9);
	GETRENDERSTATE (D3DRS_WRAP10);
	GETRENDERSTATE (D3DRS_WRAP11);
	GETRENDERSTATE (D3DRS_WRAP12);
	GETRENDERSTATE (D3DRS_WRAP13);
	GETRENDERSTATE (D3DRS_WRAP14);
	GETRENDERSTATE (D3DRS_WRAP15);
	GETRENDERSTATE (D3DRS_SEPARATEALPHABLENDENABLE);
	GETRENDERSTATE (D3DRS_SRCBLENDALPHA);
	GETRENDERSTATE (D3DRS_DESTBLENDALPHA);
	GETRENDERSTATE (D3DRS_BLENDOPALPHA);

	// load initial cached states
	for (int i = 0; i < 8; i++)
		d3d_Device->lpVtbl->GetTexture (d3d_Device, i, &d3d_StageCache[i].Texture);

	d3d_Device->lpVtbl->GetVertexDeclaration (d3d_Device, &gl_state.CurrentDecl);
	d3d_Device->lpVtbl->GetVertexShader (d3d_Device, &d3d_CurrentVS);
	d3d_Device->lpVtbl->GetPixelShader (d3d_Device, &d3d_CurrentPS);

	// get current streams...
	for (int i = 0; i < MAX_VERTEX_STREAMS; i++)
	{
		d3d_Device->lpVtbl->GetStreamSource (
			d3d_Device,
			i,
			&d3d_VertexStreams[i].pStreamData,
			&d3d_VertexStreams[i].OffsetInBytes,
			&d3d_VertexStreams[i].Stride);
	}

	// ... and indexes
	d3d_Device->lpVtbl->GetIndices (d3d_Device, &gl_state.CurrentIB);

	// set texture mode and anisotropy in diffuse and sky stages
	R_TextureMode (gl_texturemode->string, (int) gl_textureanisotropy->value);

	// set other stages
	R_SetSamplerState (TS_LIGHT, D3DTEXF_LINEAR, D3DTEXF_NONE, D3DTADDRESS_CLAMP, 1);
	R_SetSamplerState (TS_SCREEN, D3DTEXF_LINEAR, D3DTEXF_NONE, D3DTADDRESS_CLAMP, 1);
	R_SetSamplerState (TS_POSTPROC, D3DTEXF_LINEAR, D3DTEXF_NONE, D3DTADDRESS_WRAP, 1);

	// ensure that clipping is enabled
	R_SetRenderState (D3DRS_CLIPPING, FALSE);
	R_SetRenderState (D3DRS_CLIPPING, TRUE);

	// remove the alpha initially as it will be 0 on startup and until the first alpha object is drawn
	R_SetPixelShaderConstant4f (PSC_ALPHAVAL, 1, 1, 1, 1);

	// set cached state to something invalid and run a default state
	gl_state.StateBits = ~gl_state.StateBits;
	R_SetRenderStates (0);

	//d3d_Device->lpVtbl->SetSamplerState (d3d_Device, TS_DIFFUSE, D3DSAMP_SRGBTEXTURE, TRUE);
	//d3d_Device->lpVtbl->SetSamplerState (d3d_Device, TS_SKY, D3DSAMP_SRGBTEXTURE, TRUE);
	//R_SetRenderState (D3DRS_SRGBWRITEENABLE, TRUE);
}


void R_SetRenderStates (int statebits)
{
	// unchanged
	if (statebits == gl_state.StateBits) return;

	// remove depth writing from all alpha blend states
	if (statebits & RS_ALPHABLEND)
	{
		R_SetRenderState (D3DRS_BLENDOP, D3DBLENDOP_ADD);
		R_SetRenderState (D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
		R_SetRenderState (D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
		R_SetRenderState (D3DRS_ALPHABLENDENABLE, TRUE);
		statebits &= ~RS_DEPTHWRITE;
	}
	else if (statebits & RS_ADDITIVEBLEND)
	{
		R_SetRenderState (D3DRS_BLENDOP, D3DBLENDOP_ADD);
		R_SetRenderState (D3DRS_SRCBLEND, D3DBLEND_ONE);
		R_SetRenderState (D3DRS_DESTBLEND, D3DBLEND_ONE);
		R_SetRenderState (D3DRS_ALPHABLENDENABLE, TRUE);
		statebits &= ~RS_DEPTHWRITE;
	}
	else if (statebits & RS_SUBTRACTBLEND)
	{
		// a standard subtract blend is (incoming - framebuffer) but we want (framebuffer - incoming)
		R_SetRenderState (D3DRS_BLENDOP, D3DBLENDOP_REVSUBTRACT);
		R_SetRenderState (D3DRS_SRCBLEND, D3DBLEND_ONE);
		R_SetRenderState (D3DRS_DESTBLEND, D3DBLEND_ONE);
		R_SetRenderState (D3DRS_ALPHABLENDENABLE, TRUE);
		statebits &= ~RS_DEPTHWRITE;
	}
	else if (statebits & RS_MAXIMUMBLEND)
	{
		R_SetRenderState (D3DRS_BLENDOP, D3DBLENDOP_MAX);
		R_SetRenderState (D3DRS_SRCBLEND, D3DBLEND_ONE);
		R_SetRenderState (D3DRS_DESTBLEND, D3DBLEND_ONE);
		R_SetRenderState (D3DRS_ALPHABLENDENABLE, TRUE);
		statebits &= ~RS_DEPTHWRITE;
	}
	else R_SetRenderState (D3DRS_ALPHABLENDENABLE, FALSE);

	if (statebits & RS_DEPTHTEST)
	{
		R_SetRenderState (D3DRS_ZENABLE, D3DZB_TRUE);

		// it's assumed that these passes are over surfaces that have already been drawn
		if (statebits & (RS_ADDITIVEBLEND | RS_SUBTRACTBLEND | RS_MAXIMUMBLEND))
			R_SetRenderState (D3DRS_ZFUNC, D3DCMP_EQUAL);
		else R_SetRenderState (D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
	}
	else R_SetRenderState (D3DRS_ZENABLE, D3DZB_FALSE);

	if (statebits & RS_DEPTHWRITE)
		R_SetRenderState (D3DRS_ZWRITEENABLE, TRUE);
	else R_SetRenderState (D3DRS_ZWRITEENABLE, FALSE);

	if (statebits & RS_ALPHATEST)
	{
		R_SetRenderState (D3DRS_ALPHATESTENABLE, TRUE);
		R_SetRenderState (D3DRS_ALPHAREF, 1);
		R_SetRenderState (D3DRS_ALPHAFUNC, D3DCMP_GREATEREQUAL);
	}
	else R_SetRenderState (D3DRS_ALPHATESTENABLE, FALSE);

	if ((statebits & RS_CULLCCW) && gl_cull->value)
		R_SetRenderState (D3DRS_CULLMODE, D3DCULL_CCW);
	else if ((statebits & RS_CULLCW) && gl_cull->value)
		R_SetRenderState (D3DRS_CULLMODE, D3DCULL_CW);
	else R_SetRenderState (D3DRS_CULLMODE, D3DCULL_NONE);

	if (statebits & RS_WIREFRAME)
		R_SetRenderState (D3DRS_FILLMODE, D3DFILL_WIREFRAME);
	else R_SetRenderState (D3DRS_FILLMODE, D3DFILL_SOLID);

	gl_state.StateBits = statebits;
}


void R_SetEntityState (entity_t *e)
{
	// entities are always depth tested
	int statebits = RS_DEPTHTEST;

	// all entities go through here before drawing which ensures that state is consistent and correct for them all
	if (e->flags & RF_TRANSLUCENT)
		statebits |= RS_ALPHABLEND;
	else if (e->flags & RF_ADDITIVEBLEND)
		statebits |= RS_ADDITIVEBLEND;
	else if (e->flags & RF_SUBTRACTBLEND)
		statebits |= RS_SUBTRACTBLEND;
	else statebits |= RS_DEPTHWRITE;

	// set the culling mode appropriately
	if ((e->flags & RF_WEAPONMODEL) && r_lefthand->value)
		statebits |= RS_CULLCW;
	else statebits |= RS_CULLCCW;

	R_SetRenderStates (statebits);
}


void R_SetEntityTransform (glmatrix *localmatrix, int flags)
{
	// store 2 matrixes - world space and local space
	glmatrix entitymatrix;

	// derive the correct entity matrix
	if (((flags & RF_WEAPONMODEL) && r_lefthand->value) || (flags & RF_DEPTHHACK))
	{
		glmatrix flipmatrix;

		R_IdentityMatrix (&flipmatrix);

		// scale the matrix to flip from right-handed to left-handed
		if ((flags & RF_WEAPONMODEL) && r_lefthand->value) R_ScaleMatrix (&flipmatrix, -1, 1, 1);

		// scale the matrix to provide the depthrange hack (so that we don't need to set a new VP with the modified range)
		if (flags & RF_DEPTHHACK) R_ScaleMatrix (&flipmatrix, 1.0f, 1.0f, 0.3f);

		// multiply by MVP for the final matrix
		R_MultMatrix (&flipmatrix, &r_mvp_matrix, &flipmatrix);
		R_MultMatrix (&entitymatrix, localmatrix, &flipmatrix);
	}
	else R_MultMatrix (&entitymatrix, localmatrix, &r_mvp_matrix);

	R_SetVertexShaderConstant4fv (VSC_ENTITYMATRIX, entitymatrix.m16, 4);
	R_SetVertexShaderConstant4fv (VSC_LOCALMATRIX, localmatrix->m16, 4);
}


void R_SetEntityLighting (entity_t *e, float *shadelight)
{
	float an, shadevector[4];

	// PGM	ir goggles override
	if ((r_newrefdef.rdflags & RDF_IRGOGGLES) && (e->flags & RF_IR_VISIBLE))
	{
		shadelight[0] = 1.0;
		shadelight[1] = 0.0;
		shadelight[2] = 0.0;
	}

	if (r_fullbright->value)
	{
		// do it this way instead of hacking R_LightPoint so that r_lightlevel still works right
		// this doesn't work properly with all the rdflags/glows/shells/etc but r_fullbright is a cheaty hack anyway...
		shadelight[0] = 1.0;
		shadelight[1] = 1.0;
		shadelight[2] = 1.0;
	}

	an = (e->angles[0] + e->angles[1]) / 180 * M_PI;
	shadevector[0] = cos (-an);
	shadevector[1] = sin (-an);
	shadevector[2] = 1;
	Vector3Normalize (shadevector);

	R_SetPixelShaderConstant4fv (PSC_SHADELIGHT, shadelight, 1);
	R_SetPixelShaderConstant4fv (PSC_SHADEVECTOR, shadevector, 1);
}


