/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "gl_local.h"
#include "glw_win.h"

// on optimus machines the Intel is likely to be faster with Q2
// _declspec(dllexport) DWORD NvOptimusEnablement = 0x00000001;

/*
===================
QGL_WIN.C

This file implements the operating system binding of GL to QGL function
pointers.  When doing a port of Quake2 you must implement the following
two functions:

R_InitVideo() - loads libraries, assigns function pointers, etc.
R_ShutdownVideo() - unloads libraries, NULLs function pointers
===================
*/

#pragma comment (lib, "d3d9.lib")

IDirect3D9 *d3d_Object = NULL;
IDirect3DDevice9 *d3d_Device = NULL;
D3DCAPS9 d3d_DeviceCaps;
IDirect3DQuery9 *d3d_FinishEvent = NULL;

#define MAX_DEV_HANDLERS 256

// device startup/loss/reset/shutdown
typedef struct devhandler_s
{
	funcptr_t OnCreate;
	funcptr_t OnLost;
	funcptr_t OnReset;
	funcptr_t OnDestroy;
} devhandler_t;

devhandler_t d3d_DevHandlers[MAX_DEV_HANDLERS];
int d3d_NumDevHandlers = 0;

#define VH_ONCREATE		1
#define VH_ONRESET		2
#define VH_ONLOST		4
#define VH_ONDESTROY	8

static void R_RunDeviceHandlers (int hFlags)
{
	// run the specified handlers
	for (int i = 0; i < d3d_NumDevHandlers; i++)
	{
		if (hFlags & VH_ONCREATE) d3d_DevHandlers[i].OnCreate ();
		if (hFlags & VH_ONRESET) d3d_DevHandlers[i].OnReset ();
		if (hFlags & VH_ONLOST) d3d_DevHandlers[i].OnLost ();
		if (hFlags & VH_ONDESTROY) d3d_DevHandlers[i].OnDestroy ();
	}
}


// provides a dummy handler for cases where we don't want to write one ourselves
void R_NoHandler (void) {}

void R_RegisterHandler (funcptr_t OnCreate, funcptr_t OnLost, funcptr_t OnReset, funcptr_t OnDestroy)
{
	if (d3d_NumDevHandlers < MAX_DEV_HANDLERS)
	{
		d3d_DevHandlers[d3d_NumDevHandlers].OnCreate = OnCreate ? OnCreate : R_NoHandler;
		d3d_DevHandlers[d3d_NumDevHandlers].OnLost = OnLost ? OnLost : R_NoHandler;
		d3d_DevHandlers[d3d_NumDevHandlers].OnReset = OnReset ? OnReset : R_NoHandler;
		d3d_DevHandlers[d3d_NumDevHandlers].OnDestroy = OnDestroy ? OnDestroy : R_NoHandler;

		d3d_NumDevHandlers++;
	}
}


// hander registration
void Main_Register (void);
void Misc_Register (void);
void Draw_Register (void);
void Surf_Register (void);
void Part_Register (void);
void Warp_Register (void);
void Sky_Register (void);
void Mesh_Register (void);
void Null_Register (void);
void Beam_Register (void);
void Sprite_Register (void);
void BBox_Register (void);


HINSTANCE hInstD3DX = NULL;

D3DXLoadSurfaceFromMemoryProc D3DXLoadSurfaceFromMemory = NULL;
D3DXSaveSurfaceToFileProc D3DXSaveSurfaceToFile = NULL;
D3DXLoadSurfaceFromSurfaceProc D3DXLoadSurfaceFromSurface = NULL;

// default present params for modes; selected items can be changed later when the actual mode is set
static const D3DPRESENT_PARAMETERS d3d_DefaultPPWindowed = {
	0,										// backbuffer width
	0,										// backbuffer height
	D3DFMT_UNKNOWN,							// backbuffer format
	1,										// backbuffer count
	D3DMULTISAMPLE_NONE,					// no multisampling
	0,										// no multisample quality
	D3DSWAPEFFECT_DISCARD,					// discard swap
	NULL,									// no window yet; will be filled in later
	TRUE,									// windowed mode
	TRUE,									// enable depth/stencil buffer
	D3DFMT_D24S8,							// depth/stencil format
	D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL,	// discard depth/stencil too
	0,										// no refresh rate for windowed modes
	D3DPRESENT_INTERVAL_DEFAULT				// can be modified later
};

static const D3DPRESENT_PARAMETERS d3d_DefaultPPFullScreen = {
	0,										// backbuffer width will be filled in later
	0,										// backbuffer height will be filled in later
	D3DFMT_UNKNOWN,							// backbuffer format will be filled in later
	1,										// backbuffer count
	D3DMULTISAMPLE_NONE,					// no multisampling
	0,										// no multisample quality
	D3DSWAPEFFECT_DISCARD,					// discard swap
	NULL,									// no window yet; will be filled in later
	FALSE,									// fullscreen mode
	TRUE,									// enable depth/stencil buffer
	D3DFMT_D24S8,							// depth/stencil format
	D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL,	// discard depth/stencil too
	0,										// refresh rate will be filled in later
	D3DPRESENT_INTERVAL_DEFAULT				// can be modified later
};


#define MAX_DISPLAY_MODES	1024

D3DDISPLAYMODE d3d_DisplayModes[MAX_DISPLAY_MODES];
int d3d_NumDisplayModes = 0;

// for alt-enter handling
int r_lastfullscreenmode = 0;
int r_lastwindowedwidth = 0;
int r_lastwindowedheight = 0;

int R_DisplayModeCompFunc (const D3DDISPLAYMODE *a, const D3DDISPLAYMODE *b)
{
	if (a->Width == b->Width)
	{
		if (a->Height == b->Height)
		{
			if (a->RefreshRate == b->RefreshRate)
			{
				// we only enum modes for the desktop format so the formats are all the same
				return 0;
			}
			else return (int) a->RefreshRate - (int) b->RefreshRate;
		}
		else return (int) a->Height - (int) b->Height;
	}
	else return (int) a->Width - (int) b->Width;
}


qboolean R_CreateDirect3D (void)
{
	// create the D3D object
	if ((d3d_Object = Direct3DCreate9 (D3D_SDK_VERSION)) == NULL)
	{
		ri->Sys_Error (ERR_FATAL, "R_InitVideo : Direct3DCreate9 failed");
		return false;
	}

	// register our handlers
	Main_Register ();
	Misc_Register ();
	Draw_Register ();
	Surf_Register ();
	Part_Register ();
	Warp_Register ();
	Sky_Register ();
	Mesh_Register ();
	Null_Register ();
	Beam_Register ();
	Sprite_Register ();
	BBox_Register ();

	D3DDISPLAYMODE d3d_DesktopMode;

	// begin with no stored modes
	d3d_NumDisplayModes = 0;

	// get the desktop mode
	d3d_Object->lpVtbl->GetAdapterDisplayMode (d3d_Object, D3DADAPTER_DEFAULT, &d3d_DesktopMode);

	// it's 2014.  STOP DOING THIS!!!
	if (d3d_DesktopMode.Format != D3DFMT_X8R8G8B8 && d3d_DesktopMode.Format != D3DFMT_A8R8G8B8 && d3d_DesktopMode.Format != D3DFMT_A2R10G10B10)
	{
		ri->Sys_Error (ERR_FATAL, "cannot run in a non-32bpp mode");
		return false;
	}

	// get the count of modes for this format
	int ModeCount = d3d_Object->lpVtbl->GetAdapterModeCount (d3d_Object, D3DADAPTER_DEFAULT, d3d_DesktopMode.Format);

	// no modes available for this format
	if (!ModeCount)
	{
		ri->Sys_Error (ERR_FATAL, "D3D_InitModeList: Did not enumerate any modes\n");
		return false;
	}

	// set this to something safe initially
	r_lastfullscreenmode = 0;

	// enumerate them all
	for (int i = 0; i < ModeCount; i++)
	{
		// upper-bound it (done in here because while ModeCount may exceed MAX_DISPLAY_MODES, we might be
		// throwing away some of those modes, so the number of actually accepted modes may be fewer, and
		// we want to keep the high res modes at the end of the list.)
		if (d3d_NumDisplayModes >= MAX_DISPLAY_MODES) break;

		// this is the mode we're going to work on
		D3DDISPLAYMODE *mode = &d3d_DisplayModes[d3d_NumDisplayModes];

		// get the mode description for the desktop mode format
		if (FAILED (d3d_Object->lpVtbl->EnumAdapterModes (d3d_Object, D3DADAPTER_DEFAULT, d3d_DesktopMode.Format, i, mode))) break;

		// don't allow modes < 640 x 480
		if (mode->Width < 640 || mode->Height < 480) continue;

		// don't allow modes > desktop mode
		if (mode->Width > d3d_DesktopMode.Width) continue;
		if (mode->Height > d3d_DesktopMode.Height) continue;

		// if the mode width is < height we assume that we have a monitor capable of rotating it's desktop
		// and therefore we skip the mode
		if (mode->Width < mode->Height) continue;

		// accept the new mode
		d3d_NumDisplayModes++;
	}

	if (!d3d_NumDisplayModes)
	{
		ri->Sys_Error (ERR_FATAL, "D3D_InitModeList: Did not accept any enumerated modes\n");
		return false;
	}

	// sort the modes
	qsort (d3d_DisplayModes, d3d_NumDisplayModes, sizeof (D3DDISPLAYMODE), (sortfunc_t) R_DisplayModeCompFunc);

	// setup the "last" modes for alt-enter handling
	// fullscreen is the last mode in the list by default
	r_lastfullscreenmode = d3d_NumDisplayModes - 1;

	// optionally override if the desktop mode is different
	for (int i = 0; i < d3d_NumDisplayModes; i++)
	{
		D3DDISPLAYMODE *mode = &d3d_DisplayModes[i];

		// set up the lastfullscreenmode to the same as the desktop mode
		if (!memcmp (mode, &d3d_DesktopMode, sizeof (D3DDISPLAYMODE)))
		{
			r_lastfullscreenmode = i;
			break;
		}
	}

	// windowed are just mode 0 for now
	r_lastwindowedwidth = d3d_DisplayModes[0].Width;
	r_lastwindowedheight = d3d_DisplayModes[0].Height;

	// D3D was created successfully now
	return true;
}


qboolean R_GetVideoModeInfo (int *width, int *height, int mode)
{
	if (ri->vars->fullscreen->value)
	{
		// a fullscreen mode must come from the mode list and both width and height must come from the same mode
		if (mode < 0)
		{
			*width = d3d_DisplayModes[0].Width;
			*height = d3d_DisplayModes[0].Height;
		}
		else if (mode >= d3d_NumDisplayModes)
		{
			*width = d3d_DisplayModes[d3d_NumDisplayModes - 1].Width;
			*height = d3d_DisplayModes[d3d_NumDisplayModes - 1].Height;
		}
		else
		{
			*width = d3d_DisplayModes[mode].Width;
			*height = d3d_DisplayModes[mode].Height;
		}
	}
	else
	{
		// a windowed mode comes from the vid_width and vid_height cvars
		*width = ri->vars->width->value;
		*height = ri->vars->height->value;

		// lower-bound
		if (*width < d3d_DisplayModes[0].Width) *width = d3d_DisplayModes[0].Width;
		if (*height < d3d_DisplayModes[0].Height) *height = d3d_DisplayModes[0].Height;

		// upper-bound
		if (*width > d3d_DisplayModes[d3d_NumDisplayModes - 1].Width) *width = d3d_DisplayModes[d3d_NumDisplayModes - 1].Width;
		if (*height > d3d_DisplayModes[d3d_NumDisplayModes - 1].Height) *height = d3d_DisplayModes[d3d_NumDisplayModes - 1].Height;
	}

	return true;
}


D3DPRESENT_PARAMETERS *R_SetPresentParameters (void)
{
	static D3DPRESENT_PARAMETERS pp;

	if (ri->vars->fullscreen->value)
	{
		// get it from the vid_mode cvar
		D3DDISPLAYMODE *mode = &d3d_DisplayModes[(int) ri->vars->mode->value % d3d_NumDisplayModes];

		// fullscreen mode
		memcpy (&pp, &d3d_DefaultPPFullScreen, sizeof (D3DPRESENT_PARAMETERS));

		// and copy the mode to the present params
		pp.BackBufferWidth = mode->Width;
		pp.BackBufferHeight = mode->Height;
		pp.FullScreen_RefreshRateInHz = mode->RefreshRate;
		pp.BackBufferFormat = mode->Format;
	}
	else
	{
		// dimensions/etc just inherit from the parent window
		memcpy (&pp, &d3d_DefaultPPWindowed, sizeof (D3DPRESENT_PARAMETERS));

		// windowed mode (this is needed for device resets where we resize the window)
		// we can't use the client rect here as the mode hasn't yet changed!!!
		pp.BackBufferWidth = (int) ri->vars->width->value;
		pp.BackBufferHeight = (int) ri->vars->height->value;
	}

	// update the window used
	pp.hDeviceWindow = glw_state.hWnd;

	// update vsync flag
	if (ri->vars->vsync->value)
		pp.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
	else pp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;

	return &pp;
}


qboolean R_TryLoadD3DX (int version)
{
	// because we don't have d3dx any more we load it this way - yuck!!!
	if ((hInstD3DX = LoadLibrary (va ("d3dx9_%i.dll", version))) != NULL)
	{
		// now try to load them load them
		if ((D3DXLoadSurfaceFromMemory = (D3DXLoadSurfaceFromMemoryProc) GetProcAddress (hInstD3DX, "D3DXLoadSurfaceFromMemory")) == NULL) return false;
		if ((D3DXSaveSurfaceToFile = (D3DXSaveSurfaceToFileProc) GetProcAddress (hInstD3DX, "D3DXSaveSurfaceToFileA")) == NULL) return false;
		if ((D3DXLoadSurfaceFromSurface = (D3DXLoadSurfaceFromSurfaceProc) GetProcAddress (hInstD3DX, "D3DXLoadSurfaceFromSurface")) == NULL) return false;

		// loaded OK
		return true;
	}
	else
	{
		// didn't load at all
		return false;
	}
}


void R_UnloadD3DX (void)
{
	// clear the procs
	D3DXLoadSurfaceFromMemory = NULL;
	D3DXSaveSurfaceToFile = NULL;
	D3DXLoadSurfaceFromSurface = NULL;

	// and unload the library
	if (hInstD3DX)
	{
		FreeLibrary (hInstD3DX);
		hInstD3DX = NULL;
	}
}


/*
===================
R_InitVideo

This is responsible for binding our qgl function pointers to
the appropriate GL stuff.  In Windows this means doing a
LoadLibrary and a bunch of calls to GetProcAddress.  On other
operating systems we need to do the right thing, whatever that
might be.
===================
*/
qboolean R_InitVideo (HWND hWnd)
{
	// verify that all of our required objects are in the correct state
	if (!hWnd)
	{
		ri->Sys_Error (ERR_FATAL, "R_InitVideo : called with NULL hWnd");
		return false;
	}

	if (!d3d_Object)
	{
		ri->Sys_Error (ERR_FATAL, "R_InitVideo : called with NULL d3d_Object");
		return false;
	}

	if (d3d_Device)
	{
		ri->Sys_Error (ERR_FATAL, "R_InitVideo : called with non-NULL d3d_Device");
		return false;
	}

	// prep the video menu for first use by exporting the enumerated mode list and setting it up as options
	ri->PrepVideoMenu (d3d_DisplayModes, d3d_NumDisplayModes);

	// create it for first time
	if (FAILED (d3d_Object->lpVtbl->CreateDevice (
		d3d_Object,
		0,
		D3DDEVTYPE_HAL,
		hWnd,
		D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_NOWINDOWCHANGES | D3DCREATE_DISABLE_DRIVER_MANAGEMENT,
		R_SetPresentParameters (),
		&d3d_Device)))
	{
		ri->Sys_Error (ERR_FATAL, "R_InitVideo : CreateDevice failed");
		return false;
	}

	// clear to black
	d3d_Device->lpVtbl->Clear (d3d_Device, 0, NULL, D3DCLEAR_TARGET, 0, 1.0f, 0);
	d3d_Device->lpVtbl->Present (d3d_Device, NULL, NULL, NULL, NULL);

	// get caps
	d3d_Device->lpVtbl->GetDeviceCaps (d3d_Device, &d3d_DeviceCaps);

	// get the supported shader models
	const int iVertexShaderModel = D3DSHADER_VERSION_MAJOR (d3d_DeviceCaps.VertexShaderVersion);
	const int iPixelShaderModel = D3DSHADER_VERSION_MAJOR (d3d_DeviceCaps.PixelShaderVersion);

	// we're only interested in supporting SM3+ hardware
	if (iVertexShaderModel < 3 || iPixelShaderModel < 3)
	{
		ri->Sys_Error (ERR_FATAL, "R_InitVideo : iVertexShaderModel < 3 || iPixelShaderModel < 3");
		return false;
	}

	// validate other basic caps (it's assumed that the SM3 requirement will mean that all of these will pass
	if (!(d3d_DeviceCaps.TextureCaps & D3DPTEXTURECAPS_CUBEMAP))
	{
		ri->Sys_Error (ERR_FATAL, "R_InitVideo : No cubemap support");
		return false;
	}

	if (!(d3d_DeviceCaps.TextureCaps & D3DPTEXTURECAPS_MIPMAP))
	{
		ri->Sys_Error (ERR_FATAL, "R_InitVideo : No mipmap support");
		return false;
	}

	// ensure that D3DX is unloaded before we begin
	R_UnloadD3DX ();

	// starting at 99 to future-proof things a little, 23 and previous were in static libs and
	// there was no plain old "d3dx9.dll"
	for (int i = 99; i > 23; i--)
	{
		// try to load this version
		if (R_TryLoadD3DX (i)) break;

		// unload if it didn't
		R_UnloadD3DX ();
	}

	// the HINSTANCE for the library should be valid if it loaded or NULL if it didn't
	if (!hInstD3DX)
	{
		ri->Sys_Error (ERR_FATAL, "R_InitVideo : failed to load D3DX\nPlease update your installation of DirectX...");
		return false;
	}

	// setup the video dimensions from the backbuffer
	// this is done before the handlers in case anything in them depends on the dimensions
	R_SetVideoSize ();

	// setup default states
	R_SetDefaultState ();

	// make the cvars unmodified after the initial startup
	R_UnmodifyVidCvars (ri->vars);

	// run the handlers (done here in case they depend on anything that's run earlier)
	R_RunDeviceHandlers (VH_ONCREATE | VH_ONRESET);

	return true;
}


/*
===================
R_ShutdownVideo

Unloads the specified DLL then nulls out all the proc pointers.
===================
*/
void R_ShutdownVideo (void)
{
	// run the handlers
	R_RunDeviceHandlers (VH_ONLOST | VH_ONDESTROY);

	// unload D3DX
	R_UnloadD3DX ();

	// unload d3d
	SAFE_RELEASE (d3d_Device);
	SAFE_RELEASE (d3d_Object);
}


static void R_RunDeviceReset (void)
{
	// run the handlers
	R_RunDeviceHandlers (VH_ONLOST);

	// reset the device and sleep a little to give it a chance to finish
	if (FAILED (d3d_Device->lpVtbl->Reset (d3d_Device, R_SetPresentParameters ())))
	{
		ri->Sys_Error (ERR_FATAL, "Device reset failed (did you forget to Release something?)");
		return;
	}

	// spin until testcooperativelevel returns OK (in case the mode change takes some time although
	// hopefully most drivers will block until it's completed)
	while (d3d_Device->lpVtbl->TestCooperativeLevel (d3d_Device) != D3D_OK)
	{
		Sleep (10);
		ri->SendKeyEvents ();
	}
}


static void R_PostDeviceReset (void)
{
	// setup the video dimensions from the backbuffer
	// this is done before the handlers in case anything in them depends on the dimensions
	R_SetVideoSize ();

	// run the handlers
	R_RunDeviceHandlers (VH_ONRESET);

	// setup default states
	R_SetDefaultState ();
}


void R_UpdateWindowStyle (BOOL windowed)
{
	// set the styles as appropriate
	if (windowed)
		SetWindowLong (glw_state.hWnd, GWL_STYLE, WINDOW_STYLE);
	else SetWindowLong (glw_state.hWnd, GWL_STYLE, WS_POPUP);

	// and update the styles
	SetWindowPos (glw_state.hWnd, NULL, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_NOZORDER | SWP_FRAMECHANGED);

	// sleep a bit to give messages time to finish pumping around the system
	Sleep (10);

	// pump the message loop to ensure everything completes
	ri->SendKeyEvents ();
}


void R_ResetDevice (vidvars_t *vars)
{
	if (d3d_Device)
	{
		if (vars->fullscreen->value)
		{
			// remove the window styles
			if (vars->fullscreen->modified) { R_UpdateWindowStyle (FALSE); }

			// run the reset
			R_RunDeviceReset ();
		}
		else
		{
			// we can't use the client rect here because the mode hasn't changed yet
			RECT r = {0, 0, vars->width->value, vars->height->value};
			RECT workarea, cr;

			// move the window to it's new position and size
			// re-center the window after the mode change; they can change it themselves later if they want
			AdjustWindowRect (&r, WINDOW_STYLE, FALSE);

			if (SystemParametersInfo (SPI_GETWORKAREA, 0, &workarea, 0))
			{
				ri->Cvar_SetValue ("vid_xpos", workarea.left + ((workarea.right - workarea.left) - (r.right - r.left)) / 2);
				ri->Cvar_SetValue ("vid_ypos", workarea.top + ((workarea.bottom - workarea.top) - (r.bottom - r.top)) / 2);
			}

			// and now complete the move
			MoveWindow (glw_state.hWnd, vars->xpos->value, vars->ypos->value, r.right - r.left, r.bottom - r.top, TRUE);

			// run the reset
			R_RunDeviceReset ();

			// restore the window styles
			if (vars->fullscreen->modified) { R_UpdateWindowStyle (TRUE); }

			// assert the client rect
			GetClientRect (glw_state.hWnd, &cr);

			// i have an error in my code if either of these assertions are false
			assert (cr.right - cr.left == (int) ri->vars->width->value);
			assert (cr.bottom - cr.top == (int) ri->vars->height->value);

			// assert the back buffer
			IDirect3DSurface9 *bbuff = NULL;
			D3DSURFACE_DESC desc;

			if (SUCCEEDED (d3d_Device->lpVtbl->GetBackBuffer (d3d_Device, 0, 0, D3DBACKBUFFER_TYPE_MONO, &bbuff)))
			{
				if (SUCCEEDED (bbuff->lpVtbl->GetDesc (bbuff, &desc)))
				{
					// i have an error in my code if either of these assertions are false
					assert (cr.right - cr.left == desc.Width);
					assert (cr.bottom - cr.top == desc.Height);
				}

				bbuff->lpVtbl->Release (bbuff);
			}
		}

		// sleep a bit to give messages time to finish pumping around the system
		Sleep (10);

		// pump the message loop to ensure everything completes
		ri->SendKeyEvents ();

		// and finally bring everything back
		R_PostDeviceReset ();

		// and clear the modified flags in case this gets called from anywhere else
		R_UnmodifyVidCvars (ri->vars);
	}
}


void R_HandleLostDevice (void)
{
	// see if the device can be recovered
	switch (d3d_Device->lpVtbl->TestCooperativeLevel (d3d_Device))
	{
	case D3D_OK:
		// the device is no longer lost
		R_PostDeviceReset ();
		ri->Con_Printf (PRINT_ALL, "Lost device recovery complete\n");
		gl_state.DeviceLost = false;
		return;

	case D3DERR_DEVICELOST:
		// the device cannot be recovered at this time - yield some CPU time
		Sleep (10);
		break;

	case D3DERR_DEVICENOTRESET:
		// the device is ready to be reset
		R_RunDeviceReset ();
		break;

	case D3DERR_DRIVERINTERNALERROR:
	default:
		// something bad happened
		ri->Sys_Error (ERR_FATAL, "R_HandleLostDevice: D3DERR_DRIVERINTERNALERROR");
		break;
	}
}


void R_SetVideoSize (void)
{
	IDirect3DSurface9 *bbuff = NULL;
	D3DSURFACE_DESC desc;

	if (SUCCEEDED (d3d_Device->lpVtbl->GetBackBuffer (d3d_Device, 0, 0, D3DBACKBUFFER_TYPE_MONO, &bbuff)))
	{
		if (SUCCEEDED (bbuff->lpVtbl->GetDesc (bbuff, &desc)))
		{
			// set the video dimensions to the backbuffer dimensions
			vid.width = desc.Width;
			vid.height = desc.Height;

			// set up the new "last" modes for alt-enter handling
			if (ri->vars->fullscreen->value)
				r_lastfullscreenmode = ri->vars->mode->value;
			else
			{
				r_lastwindowedwidth = desc.Width;
				r_lastwindowedheight = desc.Height;
			}

			// tell the rest of the engine code about the new dimensions
			ri->Vid_NewWindow (vid.width, vid.height);
			ri->Con_Printf (PRINT_ALL, "mode %ix%i\n", desc.Width, desc.Height);
		}

		bbuff->lpVtbl->Release (bbuff);
	}
}


qboolean R_ShouldModeChange (vidvars_t *vars)
{
	// check if a vid cvar is modified and needs to trigger a mode change
	if (vars->mode->modified) return true;
	if (vars->fullscreen->modified) return true;
	if (vars->vsync->modified) return true;
	if (vars->width->modified) return true;
	if (vars->height->modified) return true;

	// update our window position
	if (vars->xpos->modified || vars->ypos->modified)
	{
		if (!vars->fullscreen->value)
		{
			RECT r;

			GetClientRect (glw_state.hWnd, &r);
			AdjustWindowRect (&r, GetWindowLong (glw_state.hWnd, GWL_STYLE), FALSE);
			MoveWindow (glw_state.hWnd, vars->xpos->value, vars->ypos->value, r.right - r.left, r.bottom - r.top, TRUE);
		}

		vars->xpos->modified = false;
		vars->ypos->modified = false;
	}

	// unmodified
	return false;
}


void R_ModifyVidCvars (vidvars_t *vars)
{
	// sets vid cvars to modified so that a mode change is triggered on the next frame
	vars->mode->modified = true;
	vars->fullscreen->modified = true;
	vars->vsync->modified = true;
	vars->width->modified = true;
	vars->height->modified = true;
}


void R_UnmodifyVidCvars (vidvars_t *vars)
{
	// sets vid cvars to unmodified so that a mode change isn't triggered on the next frame
	vars->mode->modified = false;
	vars->fullscreen->modified = false;
	vars->vsync->modified = false;
	vars->width->modified = false;
	vars->height->modified = false;
}


void R_HandleAltEnter (void)
{
	// fixme - this doesn't need to be QGL; we can move it to vid_dll and use re->vars instead
	// correction: it does need to be for the lastmode stuff
	ri->Cvar_SetValue ("vid_fullscreen", !ri->vars->fullscreen->value);

	if (ri->vars->fullscreen->value)
	{
		ri->Cvar_SetValue ("vid_mode", r_lastfullscreenmode);
	}
	else
	{
		ri->Cvar_SetValue ("vid_width", r_lastwindowedwidth);
		ri->Cvar_SetValue ("vid_height", r_lastwindowedheight);
	}
}


